/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>
             Vladimir Burian

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "qp_port.h"
#include "heating.h"
#include "dsmapi.h"
#include "selectA.h"
#include "selectB.h"
#include "util.h"
#include "datamodel.h"

//#define HEATING_ENABLED 1
//#if HEATING_ENABLED == 1

Q_DEFINE_THIS_MODULE(heating)

enum SensorOffsets {
    SENSOR_ROOM_TEMPERATURE_ACTION_OFFSET = 100,
};

enum SensorScales {
    SENSOR_TEMPERATURE_SCALE = 40,
};

enum {
    HEATING_SCALE_Yx                = 100,
    HEATING_MAX_SENSOR_AGE          = (60 * 60),    /**< the biggest allowed age of sensor values [second] */
    HEATING_SENSOR_FAILURE_TIMEOUT  = (15 * 60),    /**< timeout of tolerating invalid sensor value [seconds] */
    HEATING_CONTROL_VALUE_TIMEOUT   = (20 * 60),    /**< timeout of last sent control value [seconds] */

    HEATING_ZONE_TICK_ROUND_TICKS   = BSP_TICKS(55000), /**< length of window to process all controllers in */
};

enum {
    HEATING_INVALID_SCENE           = 0xff,
};

typedef enum {
    SENSOR_STATUS_VALID,
    SENSOR_STATUS_INVALID,
    SENSOR_STATUS_FAILURE,
} SensorStatus;

#define isInRange(x_, min_, max_)   \
    ((x_ >= min_) && (x_ <= max_))


typedef struct {
    int32_t Yp;
    int32_t Yi;
    int32_t Yd;
    int16_t Y;
    uint8_t padding[2];
} HeatingPidOutput;



static Heating l_heating;


/* Heating zone FSM --------------------------------------------------------*/
void HeatingZone_ctor(HeatingZone *me);

#define HeatingZone_init(me_)           QHsm_init((QHsm *)(me_), (QEvent *)0)
#define HeatingZone_dispatch(me_, e_)   QHsm_dispatch((QHsm *)(me_), e_)

static QState HeatingZone_initial(HeatingZone *me, QEvent const *e);
static QState HeatingZone_top(HeatingZone *me, QEvent const *e);

/* heating states */
static QState HeatingZone_internal(HeatingZone *me, QEvent const *e);
static QState HeatingZone_external(HeatingZone *me, QEvent const *e);
static QState HeatingZone_external_backup(HeatingZone *me, QEvent const *e);
static QState HeatingZone_emergency(HeatingZone *me, QEvent const *e);

/* modes of internal heating controller */
static QState HeatingZone_mode_off(HeatingZone *me, QEvent const *e);
static QState HeatingZone_mode_pid(HeatingZone *me, QEvent const *e);
static QState HeatingZone_mode_follower(HeatingZone *me, QEvent const *e);
static QState HeatingZone_mode_fixed(HeatingZone *me, QEvent const *e);
static QState HeatingZone_mode_manual(HeatingZone *me, QEvent const *e);


/* Static helper functions -------------------------------------------------*/
static void Heating_HeatingZoneLoad(Heating *me, uint8_t localZoneId, uint16_t zoneId);
static void Heating_HeatingZoneSave(Heating *me, uint8_t localZoneId);
static void Heating_HeatingZoneDispatch(Heating *me, uint8_t localZoneId, uint16_t const *zoneIds, QEvent const *e);
static bool Heating_IsSensorPushExternalControlAction(QEvent const *e);
static bool Heating_IsSensorPushExternalReferenceTemperature(QEvent const *e);
static uint8_t Heating_GetCountActiveZones(uint16_t const zoneIds[]);
static uint8_t Heating_GetNextValidLocalZoneId(uint8_t localZoneId, uint16_t const zoneIds[]);
static uint8_t Heating_GetNextFollowingZone(uint8_t localZoneId, uint16_t sourceZoneId, uint16_t const zoneIds[]);
static HeatingController *Heating_GetControllerByLocalId(uint8_t localZoneId);
static HeatingPidOutput Heating_PidProcess(HeatingController *ctrl, HeatingConfig const *config, uint16_t t, uint16_t tref, bool dry);

static void HeatingZone_Initialize(HeatingZone *me);
static void HeatingZone_PushSensor(HeatingZone *me, uint8_t sensorType, uint16_t sensorValue);
static void HeatingZone_PushSensorControl(HeatingZone *me, uint16_t sensorValue);
static void HeatingZone_OnHeatingStateChange(HeatingZone *me, uint8_t heatingState);
static void HeatingZone_OnSceneCall(HeatingZone *me, uint8_t scene);
static uint16_t HeatingZone_KeepFloorWarm(HeatingZone *me, uint16_t t, uint16_t tref, uint16_t ypid);
static SensorStatus HeatingZone_GetSensor(HeatingZone *me, uint8_t sensorType, Sensor *sensor);
static SensorStatus HeatingZone_GetSensorStatus(HeatingZone *me, uint8_t sensorType, const  Sensor *sensor);
static QStateHandler HeatingZone_GetStateByMode(uint8_t mode);
static QStateHandler HeatingZone_GetStateByControllerState(uint8_t constrollerState);

static int32_t divr(int32_t num, int32_t denum);


/*--------------------------------------------------------------------------*/
/* Heating -----------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void Heating_ctor(QActive *parent)
{
    Heating *me = &l_heating;
    uint8_t i;

    me->heatingZone.parent = parent;

    /* initialize time events */
    QTimeEvt_ctor(&me->heatingTick.tickEvt, HEATING_ZONE_MINUTE_TICK_SIG);

    /* construct heating zone FSMs */
    HeatingZone_ctor(&(me->heatingZone));
    QStateHandler state = me->heatingZone.super.state;
    for (i = 0; i < MAX_ZONES; ++i) {
        me->heatingZoneStates[i] = state;
    }
}

/*--------------------------------------------------------------------------*/
void Heating_init()
{
    Heating *me = &l_heating;
    uint8_t localZoneId;

    /* initialize heating zone FSMs */
    for (localZoneId = MIN_ZONE; localZoneId <= MAX_ZONE; ++localZoneId) {
        Heating_HeatingZoneLoad(me, localZoneId, 0);
        HeatingZone_init(&(me->heatingZone));
        Heating_HeatingZoneSave(me, localZoneId);
    }
    /* subscribe events */
    QActive_subscribe((QActive *)me->heatingZone.parent, SENSOR_PUSH_VALUE_SIG);
}

/*--------------------------------------------------------------------------*/
void Heating_dispatch(QEvent const *e, uint16_t const zoneIds[])
{
    Heating *me = &l_heating;
    switch (e->sig) {
    case HEATING_MINUTE_TICK_SIG: {
        /* get first valid zone */
        me->heatingTick.localId = Heating_GetNextValidLocalZoneId(0, zoneIds);
        if (me->heatingTick.localId) {
            /* initialize new heating tick */
            me->heatingTick.countActive = Heating_GetCountActiveZones(zoneIds);
            me->heatingTick.countActiveDispatched = 0;
            me->heatingTick.tickInterval = HEATING_ZONE_TICK_ROUND_TICKS / (me->heatingTick.countActive + 1);
            /* send minute tick to the first valid zone */
            static QEvent const tickEvt = {HEATING_ZONE_MINUTE_TICK_SIG, 0};
            QActive_postFIFO(me->heatingZone.parent, &tickEvt);
        }
        break;
    }
    case HEATING_ZONE_MINUTE_TICK_SIG: {
        uint8_t localZoneId = me->heatingTick.localId;
        /* send another minute tick if some unprocessed valid zones are left */
        me->heatingTick.localId = Heating_GetNextValidLocalZoneId(me->heatingTick.localId, zoneIds);
        if ((me->heatingTick.localId) && (++me->heatingTick.countActiveDispatched < me->heatingTick.countActive)) {
            QTimeEvt_postIn(&me->heatingTick.tickEvt, me->heatingZone.parent, me->heatingTick.tickInterval);
        }
        /* dispatch minute tick */
        Heating_HeatingZoneDispatch(me, localZoneId, zoneIds, e);
        break;
    }
    case SENSOR_PUSH_VALUE_SIG: {
        ZoneGroupSensorPushEvt const *reqEvt = (ZoneGroupSensorPushEvt const *)e;
        if ((reqEvt->localZoneId >= MIN_ZONE) && (reqEvt->localZoneId <= MAX_ZONE)) {
            Heating_HeatingZoneDispatch(me, reqEvt->localZoneId, zoneIds, e);
        }
        if ((reqEvt->zoneId > 0) &&
            (reqEvt->groupId == ROOM_TEMPERATURE_CONTROL_GROUP) &&
            (reqEvt->sensorType == SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE))
        {
            /* sensor message is room temperature action value */
            /* check if some controller is following this zone */
            uint8_t localZoneId = Heating_GetNextFollowingZone(0, reqEvt->zoneId, zoneIds);
            if (localZoneId > 0) {
                HeatingReqFollowSensorPushEvt *req = Q_NEW(HeatingReqFollowSensorPushEvt, HEATING_FOLLOW_SENSOR_PUSH_SIG);
                req->basic.localZoneId = localZoneId;
                req->zoneId = reqEvt->zoneId;
                req->sensorValue = reqEvt->sensorValue;
                QActive_postFIFO(me->heatingZone.parent, (QEvent *)req);
            }
        }
        break;
    }
    case HEATING_FOLLOW_SENSOR_PUSH_SIG: {
        HeatingReqFollowSensorPushEvt const *reqEvt = (HeatingReqFollowSensorPushEvt const *)e;
        /* dispatch signal */
        Heating_HeatingZoneDispatch(me, reqEvt->basic.localZoneId, zoneIds, e);
        /* send sensor value to the next zone in follower mode */
        uint8_t localZoneId = Heating_GetNextFollowingZone(reqEvt->basic.localZoneId, reqEvt->zoneId, zoneIds);
        if (localZoneId > 0) {
            HeatingReqFollowSensorPushEvt *req = Q_NEW(HeatingReqFollowSensorPushEvt, HEATING_FOLLOW_SENSOR_PUSH_SIG);
            req->basic.localZoneId = localZoneId;
            req->zoneId = reqEvt->zoneId;
            req->sensorValue = reqEvt->sensorValue;
            QActive_postFIFO(me->heatingZone.parent, (QEvent *)req);
        }
        break;
    }
    case CALL_SCENE_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt const *reqEvt = (SceneZoneGroupEvt const *)e;
        if ((reqEvt->groupId == ROOM_TEMPERATURE_CONTROL_GROUP) &&
            (reqEvt->basic.sceneId < HEATING_OPERATION_MODES_LENGTH))
        {
            Heating_HeatingZoneDispatch(me, reqEvt->localZoneId, zoneIds, e);
        }
        break;
    }
    case HEATING_INITIALIZE_SIG:
    case HEATING_REQ_SET_STATE_SIG:
    case HEATING_NOTIFY_CONFIG_CHANGE_SIG:
    case HEATING_NOTIFY_SCENE_CHANGE_SIG:
        /* fall through */
    case HEATING_NOTIFY_CONFIG_UPDATE_SIG: {
        HeatingReqBasicEvt const *reqEvt = (HeatingReqBasicEvt const *)e;
        Heating_HeatingZoneDispatch(me, reqEvt->localZoneId, zoneIds, e);
        break;
    }
    }
}


/*--------------------------------------------------------------------------*/
/* Heating Zone FSM --------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void HeatingZone_ctor(HeatingZone *me)
{
    QHsm_ctor(&me->super, (QStateHandler)&HeatingZone_initial);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_initial(HeatingZone *me, QEvent const *e)
{
    (void) e; //unused parameter
    memset(me->conf, 0, sizeof(HeatingConfig));
    HeatingZone_Initialize(me);
    return Q_TRAN(&HeatingZone_top);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_top(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case HEATING_INITIALIZE_SIG: {
        HeatingZone_Initialize(me);
        return Q_TRAN(&HeatingZone_internal);
    }
    case HEATING_REQ_SET_STATE_SIG: {
        HeatingReqSetState const *reqEvt = (HeatingReqSetState const *)e;
        if (reqEvt->heatingState != me->ctrl->heatingState) {
            return Q_TRAN(HeatingZone_GetStateByControllerState(reqEvt->heatingState));
        }
        return Q_HANDLED();
    }
    case HEATING_NOTIFY_CONFIG_CHANGE_SIG: {
        HeatingConfig conf;
        if (Datamodel_getHeatingConfig(me->zoneId, &conf) == ERROR_OK) {
            Heating_UpdateConfig(me->localZoneId, &conf);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_internal(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        HeatingZone_OnHeatingStateChange(me, HEATING_STATE_INTERNAL);
        return Q_HANDLED();
    }
    case Q_INIT_SIG: {
        me->ctrl->modeSuperState = (QStateHandler)&HeatingZone_internal;
        return Q_TRAN(HeatingZone_GetStateByMode(me->conf->controllerMode));
    }
    case HEATING_NOTIFY_CONFIG_UPDATE_SIG: {
        HeatingReqNotifyConfigUpdateEvt const *reqEvt = (HeatingReqNotifyConfigUpdateEvt const *)e;
        if (me->conf->controllerMode != reqEvt->prevControllerMode) {
            /* mode has changed, do transition */
            return Q_TRAN(HeatingZone_GetStateByMode(me->conf->controllerMode));
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&HeatingZone_top);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_external(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->ctrl->scene = HEATING_INVALID_SCENE;
        me->ctrl->timeOfStart = SelectB_GetTimeSeconds();
        HeatingZone_OnHeatingStateChange(me, HEATING_STATE_EXTERNAL);
        return Q_HANDLED();
    }
    case HEATING_ZONE_MINUTE_TICK_SIG: {
        if (HeatingZone_GetSensor(me, SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE, NULL) == SENSOR_STATUS_FAILURE) {
            /* control value failure */
            return Q_TRAN(&HeatingZone_external_backup);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&HeatingZone_top);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_external_backup(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        HeatingZone_OnHeatingStateChange(me, HEATING_STATE_BACKUP);
        return Q_HANDLED();
    }
    case Q_INIT_SIG: {
        me->ctrl->modeSuperState = (QStateHandler)&HeatingZone_external_backup;
        return Q_TRAN(HeatingZone_GetStateByMode(me->conf->controllerMode));
    }
    case HEATING_NOTIFY_CONFIG_UPDATE_SIG: {
        /* configuration updated, go back to external state */
        return Q_TRAN(&HeatingZone_external);
    }
    case SENSOR_PUSH_VALUE_SIG: {
        /* check whether it's possible to recover normal operation */
        if (Heating_IsSensorPushExternalControlAction(e)) {
            return Q_TRAN(&HeatingZone_external);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&HeatingZone_top);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_emergency(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        HeatingZone_OnHeatingStateChange(me, HEATING_STATE_EMERGENCY);
        return Q_HANDLED();
    }
    case HEATING_ZONE_MINUTE_TICK_SIG: {
        /* check whether it's possible to recover normal operation */
        if (me->conf->controllerMode == HEATING_MODE_PID) {
            if ((HeatingZone_GetSensor(me, SENSOR_TYPE_ROOM_TEMPERATURE, NULL) == SENSOR_STATUS_VALID) &&
                (HeatingZone_GetSensor(me, SENSOR_TYPE_ROOM_TEMPERATURE_REQUESTED, NULL) == SENSOR_STATUS_VALID))
            {
                /* temperature values are valid and up to date again */
                return Q_TRAN(&HeatingZone_internal);
            }
        }
        /* send emergency action value */
        HeatingZone_PushSensorControl(me, me->conf->emergencyValue);
        return Q_HANDLED();
    }
    case SENSOR_PUSH_VALUE_SIG: {
        /* check whether it's possible to recover normal operation */
        if ((me->conf->controllerMode == HEATING_MODE_FIXED) &&
            (Heating_IsSensorPushExternalControlAction(e)))
        {
            /* action value set manually, forget scene */
            me->ctrl->scene = HEATING_INVALID_SCENE;
            return Q_TRAN(&HeatingZone_internal);
        }
        if ((me->conf->controllerMode == HEATING_MODE_PID) &&
            (Heating_IsSensorPushExternalReferenceTemperature(e)))
        {
            /* target temperature set manually, forget scene */
            me->ctrl->scene = HEATING_INVALID_SCENE;
            return Q_HANDLED();
        }
        return Q_HANDLED();
    }
    case HEATING_FOLLOW_SENSOR_PUSH_SIG: {
        /* check whether it's possible to recover normal operation */
        if (me->conf->controllerMode == HEATING_MODE_FOLLOWER) {
            /* action value from another zone is being received again  */
            HeatingReqFollowSensorPushEvt const *reqEvt = (HeatingReqFollowSensorPushEvt const *)e;
            /* store the action value */
            me->ctrl->zfValue = reqEvt->sensorValue;
            me->ctrl->zfTimestamp = SelectB_GetTimeSeconds();
            return Q_TRAN(&HeatingZone_internal);
        }
        return Q_HANDLED();
    }
    case CALL_SCENE_ZONE_GROUP_SIG: {
        if ((me->conf->controllerMode == HEATING_MODE_PID) ||
            (me->conf->controllerMode == HEATING_MODE_FIXED))
        {
            SceneZoneGroupEvt const *reqEvt = (SceneZoneGroupEvt const *)e;
            HeatingZone_OnSceneCall(me, reqEvt->basic.sceneId);
        }
        return Q_HANDLED();
    }
    case HEATING_NOTIFY_SCENE_CHANGE_SIG: {
        /* check for scene validity */
        if (me->ctrl->scene == HEATING_INVALID_SCENE) {
            return Q_HANDLED();
        }
        /* scene is valid */
        switch (me->conf->controllerMode) {
        case HEATING_MODE_PID: {
            uint16_t operationModes[HEATING_OPERATION_MODES_LENGTH];
            if (Datamodel_getHeatingOperationModes(me->zoneId, operationModes, HEATING_OPERATION_MODES_LENGTH) == ERROR_OK) {
                HeatingZone_PushSensor(me, SENSOR_TYPE_ROOM_TEMPERATURE_REQUESTED, operationModes[me->ctrl->scene]);
            }
            return Q_HANDLED();
        }
        case HEATING_MODE_FIXED: {
            /* fixed action value is being set */
            uint16_t operationModes[HEATING_OPERATION_MODES_LENGTH];
            if (Datamodel_getHeatingOperationModes(me->zoneId, operationModes, HEATING_OPERATION_MODES_LENGTH) == ERROR_OK) {
                HeatingZone_PushSensorControl(me, operationModes[me->ctrl->scene]);
            }
            return Q_TRAN(&HeatingZone_internal);
        }
        }
        return Q_HANDLED();
    }
    case HEATING_NOTIFY_CONFIG_UPDATE_SIG: {
        /* configuration updated, go back to internal state */
        return Q_TRAN(&HeatingZone_internal);
    }
    }
    return Q_SUPER(&HeatingZone_top);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_mode_off(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->ctrl->scene = HEATING_INVALID_SCENE;
        return Q_HANDLED();
    }
    }
    return Q_SUPER(me->ctrl->modeSuperState);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_mode_pid(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->ctrl->i = 0;
        me->ctrl->e = SENSOR_INVALID_VALUE;
        me->ctrl->eprev = SENSOR_INVALID_VALUE;
        me->ctrl->timeOfStart = SelectB_GetTimeSeconds();
        me->ctrl->kfwTimestamp = SelectB_GetTimeSeconds();
        me->ctrl->kfwState = HEATING_KFW_PID;
        return Q_HANDLED();
    }
    case HEATING_ZONE_MINUTE_TICK_SIG: {
        Sensor t, tref;
        SensorStatus tStatus = HeatingZone_GetSensor(me, SENSOR_TYPE_ROOM_TEMPERATURE, &t);
        SensorStatus trefStatus = HeatingZone_GetSensor(me, SENSOR_TYPE_ROOM_TEMPERATURE_REQUESTED, &tref);
        if ((tStatus == SENSOR_STATUS_FAILURE) || (trefStatus == SENSOR_STATUS_FAILURE)) {
            /* temperature or reference temperature failure */
            return Q_TRAN(&HeatingZone_emergency);
        }
        if ((tStatus == SENSOR_STATUS_VALID) && (trefStatus == SENSOR_STATUS_VALID)) {
            /* temperatures are valid */
            /* process pid controller */
            HeatingPidOutput outPid = Heating_PidProcess(me->ctrl, me->conf, t.value, tref.value, FALSE);
            uint16_t out = outPid.Y + SENSOR_ROOM_TEMPERATURE_ACTION_OFFSET;
            /* process keep-floor-warm */
            if (me->conf->keepFloorWarm) {
                out = HeatingZone_KeepFloorWarm(me, t.value, tref.value, out);
            }
            /* send action value */
            HeatingZone_PushSensorControl(me, out);
        }
        return Q_HANDLED();
    }
    case CALL_SCENE_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt const *reqEvt = (SceneZoneGroupEvt const *)e;
        HeatingZone_OnSceneCall(me, reqEvt->basic.sceneId);
        return Q_HANDLED();
    }
    case HEATING_NOTIFY_SCENE_CHANGE_SIG: {
        if (me->ctrl->scene != HEATING_INVALID_SCENE) {
            /* scene is valid */
            uint16_t operationModes[HEATING_OPERATION_MODES_LENGTH];
            if (Datamodel_getHeatingOperationModes(me->zoneId, operationModes, HEATING_OPERATION_MODES_LENGTH) == ERROR_OK) {
                HeatingZone_PushSensor(me, SENSOR_TYPE_ROOM_TEMPERATURE_REQUESTED, operationModes[me->ctrl->scene]);
            }
        }
        return Q_HANDLED();
    }
    case SENSOR_PUSH_VALUE_SIG: {
        if (Heating_IsSensorPushExternalReferenceTemperature(e)) {
            /* target temperature set manually, forget scene */
            me->ctrl->scene = HEATING_INVALID_SCENE;
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(me->ctrl->modeSuperState);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_mode_follower(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->ctrl->scene = HEATING_INVALID_SCENE;
        me->ctrl->timeOfStart = SelectB_GetTimeSeconds();
        return Q_HANDLED();
    }
    case HEATING_FOLLOW_SENSOR_PUSH_SIG: {
        HeatingReqFollowSensorPushEvt const *reqEvt = (HeatingReqFollowSensorPushEvt const *)e;
        /* store the action value */
        me->ctrl->zfValue = reqEvt->sensorValue;
        me->ctrl->zfTimestamp = SelectB_GetTimeSeconds();
        return Q_HANDLED();
    }
    case HEATING_ZONE_MINUTE_TICK_SIG: {
        Sensor actionSensor = {
                .value = me->ctrl->zfValue,
                .timestamp = me->ctrl->zfTimestamp
        };
        /* get status of zone-follower action value */
        SensorStatus actionStatus = HeatingZone_GetSensorStatus(me, SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE, &actionSensor);
        if (actionStatus == SENSOR_STATUS_FAILURE) {
            /* action value is not being received */
            return Q_TRAN(&HeatingZone_emergency);
        }
        if (actionStatus == SENSOR_STATUS_VALID) {
            /* action value is valid */
            /* compute: actionValue = actionValue + (actionValue * offset / 100) */
            int32_t actionValue = (int32_t)me->ctrl->zfValue - SENSOR_ROOM_TEMPERATURE_ACTION_OFFSET;
            actionValue = actionValue * (me->ctrl->conf.offset + 100);
            actionValue = divr(actionValue, 100) + SENSOR_ROOM_TEMPERATURE_ACTION_OFFSET;
            /* limit action value to range 0% .. 100% */
            actionValue = (actionValue < 100) ? 100 : actionValue;
            actionValue = (actionValue > 200) ? 200 : actionValue;
            /* push the action value to the zone */
            HeatingZone_PushSensorControl(me, actionValue);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(me->ctrl->modeSuperState);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_mode_fixed(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->ctrl->timeOfStart = SelectB_GetTimeSeconds();
        return Q_HANDLED();
    }
    case HEATING_ZONE_MINUTE_TICK_SIG: {
        Sensor action;
        SensorStatus actionStatus = HeatingZone_GetSensor(me, SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE, &action);
        if (actionStatus == SENSOR_STATUS_FAILURE) {
            /* control value failure */
            return Q_TRAN(&HeatingZone_emergency);
        }
        if (actionStatus == SENSOR_STATUS_VALID) {
            /* action value is valid */
            HeatingZone_PushSensorControl(me, action.value);
        }
        return Q_HANDLED();
    }
    case CALL_SCENE_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt const *reqEvt = (SceneZoneGroupEvt const *)e;
        HeatingZone_OnSceneCall(me, reqEvt->basic.sceneId);
        return Q_HANDLED();
    }
    case HEATING_NOTIFY_SCENE_CHANGE_SIG: {
        if (me->ctrl->scene != HEATING_INVALID_SCENE) {
            /* scene is valid */
            uint16_t operationModes[HEATING_OPERATION_MODES_LENGTH];
            if (Datamodel_getHeatingOperationModes(me->zoneId, operationModes, HEATING_OPERATION_MODES_LENGTH) == ERROR_OK) {
                HeatingZone_PushSensorControl(me, operationModes[me->ctrl->scene]);
            }
        }
        return Q_HANDLED();
    }
    case SENSOR_PUSH_VALUE_SIG: {
        if (Heating_IsSensorPushExternalControlAction(e)) {
            /* action value set manually, forget scene */
            me->ctrl->scene = HEATING_INVALID_SCENE;
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(me->ctrl->modeSuperState);
}

/*--------------------------------------------------------------------------*/
static QState HeatingZone_mode_manual(HeatingZone *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->ctrl->scene = HEATING_INVALID_SCENE;
        return Q_HANDLED();
    }
    case HEATING_ZONE_MINUTE_TICK_SIG: {
        /* send manual action value */
        HeatingZone_PushSensorControl(me, me->conf->manualValue);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(me->ctrl->modeSuperState);
}


/*--------------------------------------------------------------------------*/
/* Public functions --------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void Heating_NotifyConfigChange(uint8_t localZoneId)
{
    HeatingReqBasicEvt *reqEvt = Q_NEW(HeatingReqBasicEvt, HEATING_NOTIFY_CONFIG_CHANGE_SIG);
    reqEvt->localZoneId = localZoneId;
    QActive_postFIFO(l_heating.heatingZone.parent, (QEvent *)reqEvt);
}

/*--------------------------------------------------------------------------*/
void Heating_NotifyOperationModesChange(uint8_t localZoneId)
{
    HeatingController *ctrl = Heating_GetControllerByLocalId(localZoneId);
    if (ctrl->scene != HEATING_INVALID_SCENE) {
        HeatingReqBasicEvt *reqEvt = Q_NEW(HeatingReqBasicEvt, HEATING_NOTIFY_SCENE_CHANGE_SIG);
        reqEvt->localZoneId = localZoneId;
        QActive_postFIFO(l_heating.heatingZone.parent, (QEvent *)reqEvt);
    }
}

/*--------------------------------------------------------------------------*/
void Heating_UpdateConfig(uint8_t localZoneId, HeatingConfig const *config)
{
    Heating *me = &l_heating;
    HeatingController *ctrl = Heating_GetControllerByLocalId(localZoneId);
    uint8_t prevControllerMode = ctrl->conf.controllerMode;
    /* initialize keep-floor-warm if being enabled */
    if (!ctrl->conf.keepFloorWarm && config->keepFloorWarm) {
        ctrl->kfwTimestamp = SelectB_GetTimeSeconds();
        ctrl->kfwState = HEATING_KFW_PID;
    }
    /* copy configuration to particular heating controller */
    ctrl->conf = *config;
    /* notify heating FSM about configuration update */
    HeatingReqNotifyConfigUpdateEvt *reqEvt = Q_NEW(HeatingReqNotifyConfigUpdateEvt, HEATING_NOTIFY_CONFIG_UPDATE_SIG);
    reqEvt->basic.localZoneId = localZoneId;
    reqEvt->prevControllerMode = prevControllerMode;
    QActive_postFIFO(me->heatingZone.parent, (QEvent *)reqEvt);
}

/*--------------------------------------------------------------------------*/
bool Heating_IsConfigurationValid(HeatingConfig const *config, uint8_t zoneId)
{
    if (!isInRange(config->emergencyValue, 100, 200)) {
        /* emergency value has to be always valid */
        return FALSE;
    }

    if (config->controllerMode == HEATING_MODE_OFF) {
        /* When heating is disabled, other parameters can be arbitrary */
        return TRUE;
    }

    if (config->controllerMode >= HEATING_LAST_MODE) {
        /* Unknown/unsupported heating mode */
        return FALSE;
    }

    /* check configuration of pid controller */
    if (config->controllerMode == HEATING_MODE_PID) {
        if (!isInRange(config->Kp, 0, 4000) ||
            !isInRange(config->Ti, 1, 720) ||
            !isInRange(config->Ts, 1, 60) ||
            !isInRange(config->Kd, 0, 1000) ||
            !isInRange(config->imin, -4000, 4000) ||
            !isInRange(config->imax, -4000, 4000) ||
            !isInRange(config->ymin, 0, 200) ||
            !isInRange(config->ymax, 0, 200) ||
            !isInRange(config->antiWindUp, 0, 1))
        {
            return FALSE;
        }
    }

    /* check configuration of zone follower */
    if (config->controllerMode == HEATING_MODE_FOLLOWER) {
        if ((config->sourceZoneId == 0) || (config->sourceZoneId == zoneId)) {
            return FALSE;
        }
        if (!isInRange(config->offset, -100, 300)) {
            return FALSE;
        }
    }

    /* check configuration of manual mode */
    if (config->controllerMode == HEATING_MODE_MANUAL) {
        if (!isInRange(config->manualValue, 100, 200)) {
            return FALSE;
        }
    }

    return TRUE;
}

/*--------------------------------------------------------------------------*/
bool Heating_IsHeatingStateValid(uint8_t controllerState)
{
    return (controllerState < HEATING_LAST_STATE);
}

/*--------------------------------------------------------------------------*/
bool Heating_IsControllerRunning(uint8_t localZoneId)
{
    HeatingController *ctrl = Heating_GetControllerByLocalId(localZoneId);
    return (ctrl->conf.controllerMode != HEATING_MODE_OFF);
}

/*--------------------------------------------------------------------------*/
bool Heating_GetInternals(uint8_t localZoneId, HeatingInternals *res)
{
    HeatingController *ctrl = Heating_GetControllerByLocalId(localZoneId);
    if (((ctrl->heatingState == HEATING_STATE_INTERNAL) || (ctrl->heatingState == HEATING_STATE_BACKUP)) &&
        (ctrl->conf.controllerMode == HEATING_MODE_PID) &&
        (ctrl->e != SENSOR_INVALID_VALUE))
    {
        /* PID controller is running and its input is valid */
        Sensor t = SelectA_SensorsGetSensor(localZoneId, SENSOR_TYPE_ROOM_TEMPERATURE);
        Sensor tref = SelectA_SensorsGetSensor(localZoneId, SENSOR_TYPE_ROOM_TEMPERATURE_REQUESTED);
        /* process pid controller (DRY RUN!) to get output */
        HeatingPidOutput out = Heating_PidProcess(ctrl, &ctrl->conf, t.value, tref.value, TRUE);
        /* result with internals */
        res->t = t.value;
        res->tref = tref.value;
        res->e = ctrl->e;
        res->eprev = ctrl->eprev;
        res->i = ctrl->i;
        res->Yp = out.Yp;
        res->Yi = out.Yi;
        res->Yd = out.Yd;
        res->Y = out.Y + SENSOR_ROOM_TEMPERATURE_ACTION_OFFSET;
        res->antiWindUp = ctrl->antiWindUp;
        return FALSE;
    }
    return TRUE;
}

/*--------------------------------------------------------------------------*/
uint8_t Heating_GetHeatingState(uint8_t localZoneId)
{
    HeatingController const *ctrl = Heating_GetControllerByLocalId(localZoneId);
    return ctrl->heatingState;
}

/*--------------------------------------------------------------------------*/
bool Heating_IsControllerModeOff(HeatingConfig const *config)
{
    if (!config) {
        return false;
    }
    return (config->controllerMode == HEATING_MODE_OFF);
}

/*--------------------------------------------------------------------------*/
bool Heating_initializeConfig(HeatingConfig *config)
{
    if (!config) {
        return false;
    }
    memset(config, 0, sizeof(HeatingConfig));
    return true;
}

/*--------------------------------------------------------------------------*/
bool Heating_initializeOperationModes(uint16_t *operationMode, uint16_t length)
{
    if (!operationMode) {
        return false;
    }
    memset(operationMode, 0, length * sizeof(uint16_t));
    return true;
}

/*--------------------------------------------------------------------------*/
/* Static helper functions -------------------------------------------------*/
/*--------------------------------------------------------------------------*/
static void Heating_HeatingZoneLoad(Heating *me, uint8_t localZoneId, uint16_t zoneId)
{
    Q_ASSERT((localZoneId >= MIN_ZONE) && (localZoneId <= MAX_ZONES));
    uint8_t index = localZoneId - MIN_ZONE;
    me->heatingZone.zoneId = zoneId;
    me->heatingZone.localZoneId = localZoneId;
    me->heatingZone.ctrl = &me->ctrls[index];
    me->heatingZone.conf = &me->ctrls[index].conf;
    me->heatingZone.super.state = me->heatingZoneStates[index];
}

/*--------------------------------------------------------------------------*/
static void Heating_HeatingZoneSave(Heating *me, uint8_t localZoneId)
{
    Q_ASSERT((localZoneId >= MIN_ZONE) && (localZoneId <= MAX_ZONES));
    uint8_t index = localZoneId - MIN_ZONE;
    me->heatingZoneStates[index] = me->heatingZone.super.state;
}

/*--------------------------------------------------------------------------*/
static void Heating_HeatingZoneDispatch(Heating *me, uint8_t localZoneId, uint16_t const *zoneIds, QEvent const *e)
{
    Heating_HeatingZoneLoad(me, localZoneId, zoneIds[localZoneId]);
    HeatingZone_dispatch(&(me->heatingZone), e);
    Heating_HeatingZoneSave(me, localZoneId);
}

/*--------------------------------------------------------------------------*/
static bool Heating_IsSensorPushExternalControlAction(QEvent const *e)
{
    ZoneGroupSensorPushEvt const *reqEvt = (ZoneGroupSensorPushEvt const *)e;
    if ((reqEvt->isSourceExternal) &&
        (reqEvt->groupId == ROOM_TEMPERATURE_CONTROL_GROUP) &&
        (reqEvt->sensorType == SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE))
    {
        /* sensor message is heating control action coming from external source */
        return TRUE;
    }
    return FALSE;
}

/*--------------------------------------------------------------------------*/
static bool Heating_IsSensorPushExternalReferenceTemperature(QEvent const *e)
{
    ZoneGroupSensorPushEvt const *reqEvt = (ZoneGroupSensorPushEvt const *)e;
    if ((reqEvt->isSourceExternal) &&
        (reqEvt->sensorType == SENSOR_TYPE_ROOM_TEMPERATURE_REQUESTED))
    {
        /* sensor message is reference temperature coming from external source */
        return TRUE;
    }
    return FALSE;
}

/*--------------------------------------------------------------------------*/
static uint8_t Heating_GetCountActiveZones(uint16_t const zoneIds[])
{
    uint8_t count = 0;
    uint8_t localZoneId = 0;
    while (++localZoneId <= MAX_ZONE) {
        if (zoneIds[localZoneId] !=  DATAMODEL_INVALID_ZONE_ID) {
            count++;
        }
    }
    return count;
}

/*--------------------------------------------------------------------------*/
static uint8_t Heating_GetNextValidLocalZoneId(uint8_t localZoneId, uint16_t const zoneIds[])
{
    while (++localZoneId <= MAX_ZONE) {
        if (zoneIds[localZoneId] !=  DATAMODEL_INVALID_ZONE_ID) {
            return localZoneId;
        }
    }
    return 0;
}

/*--------------------------------------------------------------------------*/
static uint8_t Heating_GetNextFollowingZone(uint8_t localZoneId, uint16_t sourceZoneId, uint16_t const zoneIds[])
{
    while (++localZoneId <= MAX_ZONE) {
        uint16_t zoneId = zoneIds[localZoneId];
        if ((zoneId != DATAMODEL_INVALID_ZONE_ID) && (zoneId != sourceZoneId)) {
            HeatingConfig const *conf = &(Heating_GetControllerByLocalId(localZoneId)->conf);
            if ((conf->controllerMode == HEATING_MODE_FOLLOWER) &&
                (conf->sourceZoneId == sourceZoneId)) {
                /* TODO ignore zones in external or emergency state */
                return localZoneId;
            }
        }
    }
    return 0;
}

/*--------------------------------------------------------------------------*/
static HeatingController *Heating_GetControllerByLocalId(uint8_t localZoneId)
{
    Q_ASSERT((localZoneId >= MIN_ZONE) && (localZoneId <= MAX_ZONE));
    return &l_heating.ctrls[localZoneId-MIN_ZONE];
}

/*--------------------------------------------------------------------------*/
static HeatingPidOutput Heating_PidProcess(HeatingController *ctrl, HeatingConfig const *config, uint16_t t, uint16_t tref, bool dry)
{
    int32_t yp, yi, yd, y, iold;
    int saturate = 0;   /* 0:not saturated; -1:negative saturation; +1:positive saturation */

    if (!dry) {
        ctrl->eprev = ctrl->e;
        ctrl->e = (int32_t)tref - (int32_t)t;

        iold = ctrl->i;
        ctrl->i += ctrl->e;

        int32_t rawimin = ((int32_t)config->imin * config->Ti) / config->Ts;
        int32_t rawimax = ((int32_t)config->imax * config->Ti) / config->Ts;

        ctrl->i = max(ctrl->i, rawimin);
        ctrl->i = min(ctrl->i, rawimax);
    }

    int16_t ymin = (int16_t)config->ymin - SENSOR_ROOM_TEMPERATURE_ACTION_OFFSET;
    int16_t ymax = (int16_t)config->ymax - SENSOR_ROOM_TEMPERATURE_ACTION_OFFSET;

    yp = divr((int32_t)HEATING_SCALE_Yx * 100 * ctrl->e, config->Kp);
    yi = divr((int32_t)ctrl->i * config->Ts, config->Ti);
    yi = divr((int32_t)HEATING_SCALE_Yx * 100 * yi, (int32_t)config->Kp);

    if (ctrl->eprev != SENSOR_INVALID_VALUE) {
        /*previous error value is valid */
        yd = divr((int32_t)HEATING_SCALE_Yx * config->Kd * (ctrl->e - ctrl->eprev), config->Ts);
    } else {
        yd = 0;
    }

    y = divr(yp+yi+yd, HEATING_SCALE_Yx);

    if (y <= -99) {
        y = -100;
        saturate = -1;
    }

    if (y >= 99) {
        y = 100;
        saturate = +1;
    }

    if ((y == 1) || (y == -1)) {
        y = 0;
    }

    if (y <= ymin) {
        y = ymin;
        saturate = -1;
    }

    if (y >= ymax) {
        y = ymax;
        saturate = +1;
    }

    if (!dry) {
        if ((config->antiWindUp) && (saturate * ctrl->e) > 0) {
            ctrl->i = iold;
            ctrl->antiWindUp = 1;
        } else {
            ctrl->antiWindUp = 0;
        }
    }

    HeatingPidOutput out;
    out.Yp = yp;
    out.Yi = yi;
    out.Yd = yd;
    out.Y = y;

    return out;
}

/*--------------------------------------------------------------------------*/
static void HeatingZone_Initialize(HeatingZone *me)
{
    me->ctrl->heatingState = HEATING_INVALID_STATE;
    me->ctrl->zfValue = SENSOR_INVALID_VALUE;
    me->ctrl->zfTimestamp = SelectB_GetTimeSeconds();
    /* Get heating configuration */
    Datamodel_getHeatingConfig(me->zoneId, me->conf);
}

/*--------------------------------------------------------------------------*/
static void HeatingZone_PushSensor(HeatingZone *me, uint8_t sensorType, uint16_t sensorValue)
{
    bool sensorPrecision = (sensorType == SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE);
    /* send sensor push to dsm-api */
    dsuid_t vdcDsuid;
    if (!(Datamodel_GetDsuidFromHeatingZone(me->zoneId, &vdcDsuid))) {
        vdsm_report(LOG_ERR, "Heating Zone Push Sensor failed to find vdc dsuid for device! Sending vdsm dsuid.\n");
        vdcDsuid = Datamodel_GetDsmid();
    }
    DsmApi_SendZoneGroupSensorPushFromVdc(&vdcDsuid, &vdcDsuid, me->zoneId, ROOM_TEMPERATURE_CONTROL_GROUP, sensorType, sensorValue, sensorPrecision);
    /* send sensor push to dS */
    ZoneGroupSensorPushEvt *reqEvt = Q_NEW(ZoneGroupSensorPushEvt, SENSOR_PUSH_VALUE_SIG);
    reqEvt->zoneId = me->zoneId;
    reqEvt->localZoneId = me->localZoneId;
    reqEvt->groupId = ROOM_TEMPERATURE_CONTROL_GROUP;
    reqEvt->sensorType = sensorType;
    reqEvt->sensorValue = sensorValue;
    reqEvt->sensorPrecision = sensorPrecision;
    reqEvt->isSourceExternal = FALSE;
    QF_publish((QEvent *)reqEvt);
}

/*--------------------------------------------------------------------------*/
static void HeatingZone_PushSensorControl(HeatingZone *me, uint16_t sensorValue)
{
    Sensor lastControl;
    uint8_t sensorType = SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE;

    lastControl = SelectA_SensorsGetSensor(me->localZoneId, sensorType);

    if ((lastControl.value != sensorValue) ||
        (SelectA_SensorsGetAge(lastControl.timestamp) >= HEATING_CONTROL_VALUE_TIMEOUT))
    {
        HeatingZone_PushSensor(me, sensorType, sensorValue);
    }
}

/*--------------------------------------------------------------------------*/
static void HeatingZone_OnHeatingStateChange(HeatingZone *me, uint8_t heatingState)
{
    me->ctrl->heatingState = heatingState;
    /* do not send dsm-api event if zone is being initialized */
    if (me->ctrl->heatingState != HEATING_INVALID_STATE) {
        DsmApi_SendHeatingControllerStateEvent(me->zoneId, heatingState);
    }
}

/*--------------------------------------------------------------------------*/
static void HeatingZone_OnSceneCall(HeatingZone *me, uint8_t scene)
{
    if (me->ctrl->scene != scene) {
        me->ctrl->scene = scene;
        HeatingReqBasicEvt *reqEvt = Q_NEW(HeatingReqBasicEvt, HEATING_NOTIFY_SCENE_CHANGE_SIG);
        reqEvt->localZoneId = me->localZoneId;
        QActive_postFIFO(me->parent, (QEvent *)reqEvt);
    }
}

/*--------------------------------------------------------------------------*/
static uint16_t HeatingZone_KeepFloorWarm(HeatingZone *me, uint16_t t, uint16_t tref, uint16_t ypid)
{
    uint32_t time = SelectB_GetTimeSeconds();

    if (t > (tref + 5*SENSOR_TEMPERATURE_SCALE)) {
        me->ctrl->kfwState = HEATING_KFW_PID;
        me->ctrl->kfwTimestamp = time;
        /* do nothing, forward PID output */
        return ypid;
    }

    uint32_t ton = 60 * (
            (t > (tref + 3*SENSOR_TEMPERATURE_SCALE)) ?  5 :   /*  5 minutes */
            (t > (tref + 1*SENSOR_TEMPERATURE_SCALE)) ? 10 :   /* 10 minutes */
                                                        15);   /* 15 minutes */
    uint32_t toff = (60*30) - ton;  /* (ton+toff) sum up to 30 minutes */

    /* check if some timing is over */
    uint32_t age = SelectA_SensorsGetAge(me->ctrl->kfwTimestamp) + 30;  /* +30s because of rounding to minutes */
    if (me->ctrl->kfwState == HEATING_KFW_FULL_OPEN) {
        if (age >= ton) {
            me->ctrl->kfwState = HEATING_KFW_PID;
            me->ctrl->kfwTimestamp = time;
        }
    } else {
        if (age >= toff) {
            me->ctrl->kfwState = HEATING_KFW_FULL_OPEN;
            me->ctrl->kfwTimestamp = time;
        }
    }

    /* keep valve open or forward PID output */
    return (me->ctrl->kfwState == HEATING_KFW_FULL_OPEN) ? me->conf->ymax : ypid;
}

/*--------------------------------------------------------------------------*/
static SensorStatus HeatingZone_GetSensor(HeatingZone *me, uint8_t sensorType, Sensor *sensor)
{
    Sensor s = SelectA_SensorsGetSensor(me->localZoneId, sensorType);

    if (sensor != NULL) {
        *sensor = s;
    }

    return HeatingZone_GetSensorStatus(me, sensorType, &s);
}

/*--------------------------------------------------------------------------*/
static SensorStatus HeatingZone_GetSensorStatus(HeatingZone *me, uint8_t sensorType, const  Sensor *sensor)
{
    /* get maximum age for the given sensor type */
    uint32_t maxSensorAge;
    switch (sensorType) {
    case SENSOR_TYPE_ROOM_TEMPERATURE:
    case SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE:
        maxSensorAge = HEATING_MAX_SENSOR_AGE;
        break;
    case SENSOR_TYPE_ROOM_TEMPERATURE_REQUESTED:
        maxSensorAge = SENSOR_MAX_AGE;
        break;
    default:
        maxSensorAge = HEATING_MAX_SENSOR_AGE;
    }

    /* deduce sensor status */
    if ((sensor->value != SENSOR_INVALID_VALUE) && (SelectA_SensorsGetAge(sensor->timestamp) <= maxSensorAge)) {
        /* sensor value is valid and up-to-date */
        return SENSOR_STATUS_VALID;
    } else if (SelectA_SensorsGetAge(me->ctrl->timeOfStart) < HEATING_SENSOR_FAILURE_TIMEOUT) {
        /* timeout to report possible sensor failure is not over yet */
        return SENSOR_STATUS_INVALID;
    } else {
        /* timeout to report failure is over */
        return SENSOR_STATUS_FAILURE;
    }
}

/*--------------------------------------------------------------------------*/
static QStateHandler HeatingZone_GetStateByMode(uint8_t mode)
{
    switch (mode) {
    case HEATING_MODE_PID:
        return (QStateHandler)&HeatingZone_mode_pid;
    case HEATING_MODE_FOLLOWER:
        return (QStateHandler)&HeatingZone_mode_follower;
    case HEATING_MODE_FIXED:
        return (QStateHandler)&HeatingZone_mode_fixed;
    case HEATING_MODE_MANUAL:
        return (QStateHandler)&HeatingZone_mode_manual;
    default:
        return (QStateHandler)&HeatingZone_mode_off;
   }
}

/*--------------------------------------------------------------------------*/
static QStateHandler HeatingZone_GetStateByControllerState(uint8_t constrollerState)
{
    switch (constrollerState) {
    case HEATING_STATE_INTERNAL:
        return (QStateHandler)&HeatingZone_internal;
    case HEATING_STATE_EXTERNAL:
        return (QStateHandler)&HeatingZone_external;
    case HEATING_STATE_BACKUP:
        return (QStateHandler)&HeatingZone_external_backup;
    case HEATING_STATE_EMERGENCY:
        return (QStateHandler)&HeatingZone_emergency;
    default:
        Q_ASSERT(FALSE);
        return 0;
    }
}

/*--------------------------------------------------------------------------*/
static int32_t divr(int32_t num, int32_t denum)
{
    int32_t absdenum = (denum >= 0) ? denum : -denum;
    int32_t numadd = absdenum / 2;
    num = num + (num>0 ? numadd : -numadd);
    return (num / denum);
}

//#endif /* HEATING_ENABLED == 1 */
