/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "selectA.h"
#include "selectB.h"
#include "registers.h"
#include "datamodel.h"
#include "ds485.h"
#include "bsp.h"
#include "group.h"
#include "light.h"
#include "security.h"
#include "access.h"
#include "heating.h"
#include "mmi.h"
#include "button.h"
#include "bitset8.h"
#include "selectA-socket.h"
#include "util.h"

#include <string.h>

Q_DEFINE_THIS_MODULE(selectA)

#define GetStateHandlerOffsetFromState(state_)  ((QStateHandler)state_)
#define GetStateFromStateHandlerOffset(offset_) (offset_)

typedef QStateHandler StateHandlerOffset;

typedef struct {
    QTimeEvt super;
    uint16_t deviceId;
    uint8_t click;
    uint8_t padding;
} ClickTimeEvt;

typedef struct {
    QEvent super;
    uint16_t deviceId;
    uint16_t zoneId;
} DevicePairingEvt;

typedef struct {
    uint8_t localZoneId;
    uint8_t groupId;
    uint8_t buttonId;
    uint8_t key;
    uint16_t fid;
    uint16_t deviceId;
    bool setLocalPriority;
    bool apartmentWide;
    uint8_t outputChannel;
} Operation;

typedef struct {
    uint8_t index;
    uint8_t count;
    uint8_t groupId[MAX_ZONE_SCENE_HISTORY];
    uint8_t scene[MAX_ZONE_SCENE_HISTORY];
} ZoneSceneHistoryRingbuffer;

typedef struct {
    uint8_t localZoneId;
    uint8_t groupId;
    uint8_t buttonGroupId;
    uint8_t padding;
    uint16_t deviceId;
} RegisterLocalPrio;

typedef struct {
    uint32_t timestamps[MAX_ZONES][MAX_CACHED_SENSORS]; /**< time of the last change */
    uint16_t values[MAX_ZONES][MAX_CACHED_SENSORS];     /**< values in 10.bit format */
    uint8_t padding[2];
} Sensors;

#define ZONE_HIST_FLAG_FORCE  0x40
#define ZONE_HIST_FLAG_UNDO   0x80

enum {
    SELECT_A_RESEND_TIMEOUT_TICKS   =   BSP_TICKS(2000),
};

/* Active object class -----------------------------------------------------*/
typedef struct {
    QActive super;
    QEQueue deferredQueue;              /**< deferred event queue */
    SharedOpTimeEvt sharedTimeEvts1[8]; /**< shared first time events used for all state machines */
    SharedOpTimeEvt sharedTimeEvts2[8]; /**< shared second time events used for all state machines */
    ClickTimeEvt clickTimeEvts[8];      /**< click time events used for detecting resends (ARQ) */
    StopTimerEvt stopTimeEvts[20];      /**< timers for STOP timeout handling */
    Sensors sensors;                    /**< cache for sensor values */
    uint16_t select;                    /**< current A selection */
    uint16_t source;                    /**< short upstream telegram source address */
    uint8_t key;                        /**< short upstream telegram key ID */
    uint8_t click;                      /**< short upstream telegram click type */
    uint8_t groupId;                    /**< short upstream button group ID */
    uint8_t apartmentId;                /**< cached apartment ID */
    uint8_t ds485Address;               /**< cached dS485 device address */
    uint8_t transactionId;              /**< last transaction ID */
    uint8_t actionMode;                 /**< action mode (direct scene calls) */
    uint8_t actionId;                  /**< action Id (direct scene calls) */
    ShortUpstreamEvt lastUpstreamEvt;   /**< cache upstream event until device is available */
    uint32_t binaryInputEventCount;     /**< binary input event count */
    DatamodelDevice device;             /**< data model device structure */
    ProgrammingMode progMode;           /**< programming mode structure */
    Light light;                        /**< shared light extended state used for any group */
    Security security;                  /**< shared security extended state used for any group */
    Access access;                      /**< shared access extended state used for any group */
    Group group;                        /**< shared group extended state */
    StateHandlerOffset lightHsmStateOffsets[NUM_STANDARD_GROUPS][MAX_ZONES];     /**< current state handler RAM offset address of every light main group */
    StateHandlerOffset buttonGroupHsmStateOffsets[NUM_GROUPS_WITH_AREA][MAX_ZONES][4];   /**< current state handler RAM offset address of every light button group */
    /* State handlers for all user group HSMs */
    StateHandlerOffset clusterHsmStateOffsets[NUM_USER_GROUP_WITH_HSM];
    StateHandlerOffset lightGroupInitHsmStateOffset;
    ZoneSceneHistoryRingbuffer zoneSceneHistoryRingbuffer[MAX_ZONES];
    uint16_t zoneIds[MAX_ZONE+2];       /**< cached zone IDs */
    uint16_t pairingDeviceId;
    uint16_t pairingZoneId;
    bool pairingZoneCreated;
    ApartmentState_t apartmentState;    /**< state of the apartment */
    RegisterLocalPrio registerLocalPrios[4];
    Scene alarmScene;                   /**< cache current alarm scene */
    uint8_t stateMachineGroupId[NUM_USER_GROUP_WITH_HSM];    /**< cached statemachine groupId's for usergroups with statemachine */
    uint8_t padding;
} SelectA;

/* Local objects -----------------------------------------------------------*/
/* the single instance of the selectA active object */
static SelectA l_selectA;
/* event queue storage */
static QEvent const * l_deferredQueueSto[8];
/* key (click type) to signal table for any button except a local button */
static const QSignal l_keyToSignal[] = {
    CT_TIP_1X_SIG,
    CT_TIP_2X_SIG,
    CT_TIP_3X_SIG,
    CT_TIP_4X_SIG,
    CT_HOLD_START_SIG,
    CT_HOLD_REPEAT_SIG,
    CT_HOLD_END_SIG,
    CT_CLICK_1X_SIG,
    CT_CLICK_2X_SIG,
    CT_CLICK_3X_SIG,
    CT_SHORT_LONG_SIG,
    CT_LOCAL_OFF_SIG,
    CT_LOCAL_ON_SIG,
};
/* key (click type) to signal table for 2-way-up button */
static const QSignal l_2WayUpButtonKeyToSignal[] = {
    CT_2WAY_UP_TIP_1X_SIG,
    CT_2WAY_UP_TIP_2X_SIG,
    CT_2WAY_UP_TIP_3X_SIG,
    CT_2WAY_UP_TIP_4X_SIG,
    CT_2WAY_UP_HOLD_START_SIG,
    CT_2WAY_UP_HOLD_REPEAT_SIG,
    CT_2WAY_UP_HOLD_END_SIG,
    CT_2WAY_UP_TIP_1X_SIG,
    CT_2WAY_UP_TIP_2X_SIG,
    CT_2WAY_UP_TIP_3X_SIG,
    CT_SHORT_LONG_SIG,
    CT_LOCAL_OFF_SIG,
    CT_LOCAL_ON_SIG,
};
/* key (click type) to signal table for 2-way-down button */
static const QSignal l_2WayDownButtonKeyToSignal[] = {
    CT_2WAY_DOWN_TIP_1X_SIG,
    CT_2WAY_DOWN_TIP_2X_SIG,
    CT_2WAY_DOWN_TIP_3X_SIG,
    CT_2WAY_DOWN_TIP_4X_SIG,
    CT_2WAY_DOWN_HOLD_START_SIG,
    CT_2WAY_DOWN_HOLD_REPEAT_SIG,
    CT_2WAY_DOWN_HOLD_END_SIG,
    CT_2WAY_DOWN_TIP_1X_SIG,
    CT_2WAY_DOWN_TIP_2X_SIG,
    CT_2WAY_DOWN_CLICK_3X_SIG,
    CT_SHORT_LONG_SIG,
    CT_LOCAL_OFF_SIG,
    CT_LOCAL_ON_SIG,
};
/* key (click type) to signal table for local 2-way-down button (SDS) */
static const QSignal l_local2WayDownButtonKeyToSignal[] = {
    CT_2WAY_DOWN_TIP_1X_SIG,
    CT_2WAY_DOWN_TIP_2X_SIG,
    CT_2WAY_DOWN_TIP_3X_SIG,
    CT_2WAY_DOWN_TIP_4X_SIG,
    CT_LOCAL_2WAY_DOWN_HOLD_START_SIG,
    CT_LOCAL_2WAY_DOWN_HOLD_REPEAT_SIG,
    CT_LOCAL_2WAY_DOWN_HOLD_END_SIG,
    CT_2WAY_DOWN_TIP_1X_SIG,
    CT_2WAY_DOWN_TIP_2X_SIG,
    CT_2WAY_DOWN_CLICK_3X_SIG,
    CT_SHORT_LONG_SIG,
    CT_LOCAL_OFF_SIG,
    CT_LOCAL_ON_SIG,
};
/* key (click type) to signal table for a local button */
static const QSignal l_localKeyToSignal[] = {
    CT_TIP_1X_SIG,
    CT_TIP_2X_SIG,
    CT_TIP_3X_SIG,
    CT_TIP_4X_SIG,
    CT_LOCAL_HOLD_START_SIG,
    CT_LOCAL_HOLD_REPEAT_SIG,
    CT_LOCAL_HOLD_END_SIG,
    CT_CLICK_1X_SIG,
    CT_CLICK_2X_SIG,
    CT_CLICK_3X_SIG,
    CT_LOCAL_SHORT_LONG_SIG,
    CT_LOCAL_OFF_SIG,
    CT_LOCAL_ON_SIG,
    CT_LOCAL_SHORT_SHORT_LONG_SIG,
    CT_LOCAL_STOP_SIG,
};

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_SelectA = (QActive *)&l_selectA;      /* "opaque" AO pointer */

/* Local methods -----------------------------------------------------------*/
static QState SelectA_initial(SelectA *me, QEvent const *e);
static QState SelectA_top(SelectA *me, QEvent const *e);
static QState SelectA_wait_for_configuration(SelectA *me, QEvent const *e);
static QState SelectA_idle(SelectA *me, QEvent const *e);
static QState SelectA_busy(SelectA *me, QEvent const *e);
static QState SelectA_retrieve_device_for_short_upstream(SelectA *me, QEvent const *e);
static QState SelectA_device_set_zone(SelectA *me, QEvent const *e);
static QState SelectA_device_create_zone(SelectA *me, QEvent const *e);

/* helper functions */
static SharedOpTimeEvt * SelectA_GetSharedOpTimeEvt1(SelectA *me, uint8_t localZoneId, uint8_t groupId, uint8_t buttonId, uint8_t outputChannel);
static SharedOpTimeEvt * SelectA_GetSharedOpTimeEvt2(SelectA *me, uint8_t localZoneId, uint8_t groupId, uint8_t buttonId, uint8_t outputChannel);
static StopTimerEvt * SelectA_GetFreeStopTimer(SelectA *me);
static void SelectA_UnhandledZoneGroupOperation(QEvent const *e);
static void SelectA_HandleOperation(SelectA *me, Operation const * op, QEvent const *e);
static void SelectA_HandleShortUpstreamOperation(SelectA *me);
static void SelectA_HandleSceneAllOperation(SelectA *me, SceneZoneGroupEvt const * e);
static void SelectA_HandleSceneGroupOperation(SelectA *me, SceneZoneGroupEvt const * e);
static void SelectA_HandleSceneZoneOperation(SelectA *me, SceneZoneGroupEvt const * e);
static void SelectA_HandleSceneZoneGroupOperation(SelectA *me, SceneZoneGroupEvt const * e);
static void SelectA_HandleSceneDeviceOperation(SelectA *me, SceneDeviceEvt const * e);
static void SelectA_HandleSetLocalPrioAllOperation(SelectA *me, ActionZoneGroupSceneEvt const * e);
static void SelectA_HandleSetLocalPrioGroupOperation(SelectA *me, ActionZoneGroupSceneEvt const * e);
static void SelectA_HandleSetLocalPrioZoneOperation(SelectA *me, ActionZoneGroupSceneEvt const * e);
static void SelectA_HandleSetLocalPrioZoneGroupOperation(SelectA *me, ActionZoneGroupSceneEvt const * e);
static void SelectA_HandleCallSceneMinAllOperation(SelectA *me, ActionZoneGroupSceneEvt const * e);
static void SelectA_HandleRegisterLocalPrioZoneOperation(SelectA *me, ActionZoneGroupSceneEvt const * e);
static void SelectA_HandleRegisterLocalPrioZoneGroupOperation(SelectA *me, ActionZoneGroupSceneEvt const * e);
static void SelectA_HandleSharedTimer1Operation(SelectA *me, SharedOpTimeEvt const *opTimeEvt);
static void SelectA_HandleSharedTimer2Operation(SelectA *me, SharedOpTimeEvt const *opTimeEvt);
static void SelectA_HandleProgrammingModeSignal(SelectA *me, QEvent const * e);
static void SelectA_HandleDatamodelDeviceEntryResponse(SelectA *me, DatamodelResDeviceEntryEvt const *deviceEntryEvt);
static uint8_t SelectA_GetTransactionId(SelectA *me);
static void SelectA_InvalidateCaches(SelectA *me);
static void SelectA_UpdateUserGroup(SelectA *me, uint8_t groupId, uint8_t stateMachineGroupId);
static void SelectA_TxOPCOperation(SelectA *me, OPCOperation operation, uint8_t channelId);
static bool SelectA_CheckStatelessScene(Scene scene);
static void SelectA_SaveCallSceneZoneGroupHistory(SelectA * selectA, uint8_t localZoneId, uint8_t groupId, Scene scene, uint8_t flags);
static uint8_t SelectA_GetGroupId(DatamodelDevice const *device, uint8_t key);
static uint8_t SelectA_GetButtonId(DatamodelDevice const *device, uint8_t key);
static bool SelectA_GetLocalPriority(DatamodelDevice const *device, uint8_t key);
static uint16_t SelectA_GetFunctionId(DatamodelDevice const *device);
static uint8_t SelectA_GetLocalZoneId(SelectA *me, uint16_t zoneId);
static bool SelectA_HandleResendFlag(SelectA *me);
static void SelectA_HandleShortUpstream(SelectA *me);
static StopTimerEvt * SelectA_GetStopTimer(uint8_t localZoneId, uint8_t groupId, uint8_t buttonId);

QSignal SelectA_GetSignal(uint8_t clickType, const Operation *op);
static void SelectA_SensorsInit(Sensors *sensors);
static void SelectA_SensorsInitZone(Sensors *sensors, uint8_t localZoneId);
static void SelectA_SensorsUpdateSensor(ZoneGroupSensorPushEvt const *reqEvt);
static uint8_t SelectA_SensorsGetSensorIdx(uint8_t sensorType, uint8_t groupId);

/*..........................................................................*/
void SelectA_CallSceneZoneGroupApartment(uint8_t localZoneId, uint8_t groupId, uint8_t sourceZoneId, uint8_t sourceGroupId, Scene scene, bool apartmentWide, uint16_t originDevice)
{
    SelectA_CallSceneZoneGroup(apartmentWide ? 0 : localZoneId, groupId, sourceZoneId, sourceGroupId, scene, originDevice);
}

/*..........................................................................*/
uint8_t SelectA_CallSceneZoneGroup(uint8_t localZoneId, uint8_t groupId, uint8_t sourceZoneId, uint8_t sourceGroupId, Scene scene, uint16_t originDevice)
{
    /* remote: send call scene dSM API call */
    DsmApi_SendZoneGroupActionCallSceneRequest(localZoneId, groupId, scene, originDevice, FALSE);
    /* local */
    if ((localZoneId != 0) && (groupId != 0)) {
        /* zone/group scene */
        /* send call scene to line */
        return SelectA_SendSceneZoneGroup(localZoneId, groupId, scene, FALSE);
    } else {
        SceneZoneGroupEvt * e = Q_NEW(SceneZoneGroupEvt, CALL_SCENE_BROADCAST_SIG);
        e->basic.sceneId = scene;
        e->basic.sourceZoneId = sourceZoneId;
        e->basic.sourceGroupId = sourceGroupId;
        e->basic.force = FALSE;
        e->basic.outputChannel = 0;
        e->localZoneId = localZoneId;
        e->groupId = groupId;
        QActive_postLIFO((QActive *)&l_selectA, (QEvent *)e);
    }
    return SELECTA_NO_DOWNSTREAM_MESSAGE;
}

/*..........................................................................*/
void SelectA_ForceCallSceneZoneGroupApartment(uint8_t localZoneId, uint8_t groupId, uint8_t sourceZoneId, uint8_t sourceGroupId, Scene scene, bool apartmentWide, uint16_t originDevice)
{
    SelectA_ForceCallSceneZoneGroup(apartmentWide? 0 : localZoneId, groupId, sourceZoneId, sourceGroupId, scene, originDevice);
}

/*..........................................................................*/
void SelectA_ForceCallSceneZoneGroup(uint8_t localZoneId, uint8_t groupId, uint8_t sourceZoneId, uint8_t sourceGroupId, Scene scene, uint16_t originDevice)
{
    /* remote: send force call scene dSM API call */
    DsmApi_SendZoneGroupActionCallSceneRequest(localZoneId, groupId, scene, originDevice, TRUE);
    /* local */
    if ((localZoneId != 0) && (groupId != 0)) {
        /* zone/group scene */
        /* send force call scene to line */
        SelectA_SendSceneZoneGroup(localZoneId, groupId, scene, TRUE);
    } else {
        SceneZoneGroupEvt * e = Q_NEW(SceneZoneGroupEvt, CALL_SCENE_BROADCAST_SIG);
        e->basic.sceneId = scene;
        e->basic.sourceZoneId = sourceZoneId;
        e->basic.sourceGroupId = sourceGroupId;
        e->basic.force = TRUE;
        e->basic.outputChannel = 0;
        e->localZoneId = localZoneId;
        e->groupId = groupId;
        QActive_postLIFO((QActive *)&l_selectA, (QEvent *)e);
    }
}

/*..........................................................................*/
uint8_t SelectA_SendSceneZoneGroup(uint8_t localZoneId, uint8_t groupId, Scene scene, bool force)
{
    uint8_t numShortDSMsgs = SELECTA_LONG_DOWNSTREAM_MESSAGE;
    SelectASocket_SendSceneZoneGroup(localZoneId, groupId, scene, force);
    if (SelectA_CheckStatelessScene(scene)) {
        /* save call scene into history */
        SelectA_SaveCallSceneZoneGroupHistory(&l_selectA, localZoneId, groupId, scene, (force) ? ZONE_HIST_FLAG_FORCE : 0);
    }
    return numShortDSMsgs;
}

/*..........................................................................*/
void SelectA_SendSceneDevice(uint16_t deviceId, Scene scene)
{
    SelectASocket_SendSceneDevice(deviceId, scene, FALSE);
}

/*..........................................................................*/
void SelectA_UndoSceneNumberBroadcast(uint8_t sourceZoneId, uint8_t sourceGroupId, Scene scene, uint16_t originDevice)
{
    SceneZoneGroupEvt * e = Q_NEW(SceneZoneGroupEvt, UNDO_SCENE_NUMBER_BROADCAST_SIG);
    e->basic.sceneId = scene;
    e->basic.sourceZoneId = sourceZoneId;
    e->basic.sourceGroupId = sourceGroupId;
    e->localZoneId = 0;
    e->groupId = 0;
    e->basic.force = FALSE;
    e->basic.outputChannel = 0;
    QActive_postLIFO((QActive *)&l_selectA, (QEvent *)e);
    /* undo scene number remotely */
    DsmApi_SendZoneGroupActionUndoSceneNumberRequest(0, 0, scene, originDevice);
}

/*..........................................................................*/
void SelectA_SetSceneLocalPrioZoneGroup(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice)
{
    Q_ASSERT((localZoneId != 0) && (groupId != 0));
    /* set local priority on scene locally */
    SelectASocket_SendSetSceneLocalPrioZoneGroup(localZoneId, groupId, scene);
    /* set local priority on scene remotely */
    DsmApi_SendZoneGroupActionSetSceneLocalPrioRequest(localZoneId, groupId, scene, originDevice);
}

/*..........................................................................*/
uint8_t SelectA_CallSceneMinZoneGroup(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice)
{
    Q_ASSERT(groupId != 0);
    uint8_t numShortDSMsgs = SELECTA_LONG_DOWNSTREAM_MESSAGE;
    /* call min scene */
    SelectASocket_CallSceneMinZoneGroup(localZoneId, groupId, scene);
    /* call min scene remotely */
    DsmApi_SendZoneGroupActionCallSceneMinRequest(localZoneId, groupId, scene, originDevice);
    return numShortDSMsgs;
}

/*..........................................................................*/
void SelectA_ProgOnZoneGroup(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice)
{
    Q_ASSERT((localZoneId != 0) && (groupId != 0));
    /* set local priority on scene locally */
    SelectA_SendProgOnZoneGroup(localZoneId, groupId);
    /* send PROG_ON remotely */
    DsmApi_SendZoneGroupActionProgOn(localZoneId, groupId, originDevice);
}

/*..........................................................................*/
void SelectA_ProgOffZoneGroup(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice)
{
    Q_ASSERT((localZoneId != 0) && (groupId != 0));
    /* set local priority on scene locally */
    SelectA_SendProgOffZoneGroup(localZoneId, groupId);
    /* send PROG_ON remotely */
    DsmApi_SendZoneGroupActionProgOff(localZoneId, groupId, originDevice);
}

/*..........................................................................*/
void SelectA_SaveSceneZoneGroup(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice)
{
    Q_ASSERT((localZoneId != 0) && (groupId != 0));
    /* set local priority on scene locally */
    SelectA_SendSaveSceneZoneGroup(localZoneId, groupId, scene);
    /* send SaveScene remotely */
    DsmApi_SendZoneGroupActionSaveScene(localZoneId, groupId, scene, originDevice);
}

/*..........................................................................*/
void SelectA_CallOPCDimZoneGroup(uint8_t localZoneId, uint8_t groupId, OPCOperation operation, uint8_t channelId, uint16_t originDevice)
{
    /* call OPC command */
    SelectA_SendOPCDimZoneGroup(localZoneId, groupId, operation, channelId);
    /* call OPC command remotely */
    DsmApi_SendZoneGroupActionOPCDimRequest(localZoneId, groupId, operation, channelId, originDevice);
}

/*..........................................................................*/
void SelectA_SendUndoSceneNumberZoneGroup(uint8_t localZoneId, uint8_t groupId, Scene scene)
{
    SelectASocket_SendUndoSceneNumberZoneGroup(localZoneId, groupId, scene);
    if (SelectA_CheckStatelessScene(scene)) {
        /* save call scene into history */
        SelectA_SaveCallSceneZoneGroupHistory(&l_selectA, localZoneId, groupId, scene, ZONE_HIST_FLAG_UNDO);
    }
}

/*..........................................................................*/
void SelectA_SendBlinkZoneGroup(uint8_t localZoneId, uint8_t groupId)
{
    SelectASocket_SendBlinkZoneGroup(l_selectA.zoneIds[localZoneId], groupId);
}

/*..........................................................................*/
void SelectA_SendProgOnZoneGroup(uint8_t localZoneId, uint8_t groupId)
{
    SelectASocket_SendProgOnZoneGroup(l_selectA.zoneIds[localZoneId], groupId);
}

/*..........................................................................*/
void SelectA_SendProgOnDevice(uint16_t deviceId)
{
    SelectASocket_SendProgOnDevice(deviceId);
}

/*..........................................................................*/
void SelectA_SendProgOffZoneGroup(uint8_t localZoneId, uint8_t groupId)
{
    SelectASocket_SendProgOffZoneGroup(l_selectA.zoneIds[localZoneId], groupId);
}

/*..........................................................................*/
void SelectA_SendProgOffDevice(uint16_t deviceId)
{
    SelectASocket_SendProgOffDevice(deviceId);
}

/*..........................................................................*/
void SelectA_SendSaveSceneZoneGroup(uint8_t localZoneId, uint8_t groupId, Scene scene)
{
    SelectASocket_SendSaveSceneZoneGroup(localZoneId, groupId, scene);
}

/*..........................................................................*/
void SelectA_SendOPCDimZoneGroup(uint8_t localZoneId, uint8_t groupId, OPCOperation operation, uint8_t channelId)
{
   SelectASocket_SendOPCDimZoneGroup(localZoneId, groupId, operation, channelId);
}

/*..........................................................................*/
void SelectA_SendOPCDimDevice(uint16_t deviceId, OPCOperation operation, uint8_t channelId)
{
    SelectASocket_SendOPCDimDevice(deviceId, operation, channelId);
}

/*..........................................................................*/
void SelectA_SendGeneralAck(void)
{
    /* NOP */
}

/*..........................................................................*/
uint32_t SelectA_GetBinaryInputEventCount(void)
{
    SelectA *me = &l_selectA;
    return me->binaryInputEventCount;
}

/*..........................................................................*/

enum {
  ACM_NORMAL = 0,
  ACM_FORCE  = 1,
  ACM_UNDO   = 2
};

static void initializeSceneZoneGroupEvent(SceneZoneGroupEvt *e, SelectA *me)
{
    Q_ASSERT (e);
    e->basic.sourceZoneId = 0;
    e->basic.sourceGroupId = 0;
    e->basic.outputChannel = 0;
    e->basic.sourceDeviceId = me->device.deviceId;
    e->basic.sceneId = me->actionId;
    e->basic.force = FALSE;
    e->groupId = me->groupId;
    e->localZoneId = me->device.localZoneId;
}

void SelectA_DirectSceneCall(SelectA *me)
{
    SceneZoneGroupEvt *e = NULL;

    bool callApartment = ((me->actionId >= 64) &&
                          (me->actionId != 67) &&  // Room Standby
                          (me->actionId != 68) &&  // Room deep off
                          (me->actionId != 75));   // Zone Active

    if ((me->actionMode == ACM_NORMAL) ||
        (me->actionMode == ACM_FORCE)) {
        if (callApartment) {
            /* send broadcast to apartment */
            e = Q_NEW(SceneZoneGroupEvt, CALL_SCENE_BROADCAST_SIG);
            initializeSceneZoneGroupEvent(e, me);
            e->groupId = 0;
            e->localZoneId = 0;
        } else {
            e = Q_NEW(SceneZoneGroupEvt, CALL_SCENE_ZONE_GROUP_SIG);
            initializeSceneZoneGroupEvent(e, me);
            if (me->actionId >= 64) {
                e->groupId = 0;
            }
        }
        e->basic.force = (me->actionMode == ACM_FORCE);
        DsmApi_SendZoneGroupActionCallSceneRequest(e->localZoneId, e->groupId, e->basic.sceneId, e->basic.sourceDeviceId, e->basic.force);
    } else if (me->actionMode == ACM_UNDO) {
        if (callApartment) {
            /* send broadcast to apartment */
            e = Q_NEW(SceneZoneGroupEvt, UNDO_SCENE_NUMBER_BROADCAST_SIG);
            initializeSceneZoneGroupEvent(e, me);
            e->groupId = 0;
            e->localZoneId = 0;
            e->basic.force = TRUE;
        } else {
            e = Q_NEW(SceneZoneGroupEvt, UNDO_SCENE_NUMBER_ZONE_GROUP_SIG);
            initializeSceneZoneGroupEvent(e, me);
            if (me->actionId >= 64) {
                e->groupId = 0;
            }
            e->basic.force = TRUE;
        }
        DsmApi_SendZoneGroupActionUndoSceneNumberRequest(e->localZoneId, e->groupId, e->basic.sceneId, e->basic.sourceDeviceId);
    } else {
        vdsm_report(LOG_WARNING, "SelectA_DirectSceneCall action mode value: 0x%04x not supported\n", me->actionMode);
        return;
    }

    QActive_postLIFO(AO_SelectA, (QEvent *)e);
}

/*..........................................................................*/
void SelectA_PostGroupSignal(uint8_t groupId, uint8_t localZoneId, QSignal signal, bool force)
{
    SelectA *me = &l_selectA;
    Operation op;
    ClickForwardEvt newEvent;
    /* setup event */
    newEvent.super.sig = signal;
    newEvent.super.dynamic_ = 0;
    newEvent.groupId = groupId;
    newEvent.localZoneId = localZoneId;
    newEvent.key = 0;
    newEvent.fid = 0;
    newEvent.force = force;
    /* setup operation */
    op.groupId = groupId;
    op.localZoneId = localZoneId;
    op.key = 0;
    op.fid = 0;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = 0;
    for (op.buttonId = BUTTON_GROUP_1; op.buttonId <= BUTTON_GROUP_4; ++op.buttonId) {
        SelectA_HandleOperation(me, &op, (QEvent *)&newEvent);
    }
}

/*..........................................................................*/
static bool SelectA_HandleResendFlag(SelectA *me)
{
    ShortUpstreamEvt const * shortUpEvt = &me->lastUpstreamEvt;
    uint16_t deviceId = shortUpEvt->source;
    uint8_t click = (shortUpEvt->sensor << 7) | shortUpEvt->click;
    bool exit = FALSE;
    uint8_t i;
    /* do we have a running timer for this device? */
    for (i = 0; i < Q_DIM(me->clickTimeEvts); ++i) {
        ClickTimeEvt * clickTimeEvt = &me->clickTimeEvts[i];
        if (clickTimeEvt->deviceId == deviceId) {
            /* we have a running timer for this device */
            /* stop timer */
            if (QTimeEvt_disarm(&(clickTimeEvt->super))) {
                /* timer could be stopped -> clear device ID */
                clickTimeEvt->deviceId = 0;
            } else {
                /* timer could not be stopped -> just invalidate device ID */
                clickTimeEvt->deviceId = MAX_DEVICE + 1;
            }
            /* is the packet a re-send (ARQ)? */
            if (shortUpEvt->resend) {
                /* is the click identical? */
                if (clickTimeEvt->click == click) {
                    /* don't handle this click (it is a re-send of a click we already handled), just exit */
                    exit = TRUE;
                }
            }
            break;
        }
    }
    /* start first free click timer */
    for (i = 0; i < Q_DIM(me->clickTimeEvts); ++i) {
        ClickTimeEvt * clickTimeEvt = &me->clickTimeEvts[i];
        if (clickTimeEvt->deviceId == 0) {
            clickTimeEvt->deviceId = deviceId;
            clickTimeEvt->click = click;
            QTimeEvt_postIn(&(clickTimeEvt->super), (QActive *)me, SELECT_A_RESEND_TIMEOUT_TICKS);
            break;
        }
    }
    return exit;
}

/*..........................................................................*/
static void SelectA_HandleShortUpstream(SelectA *me)
{
    ShortUpstreamEvt const * shortUpEvt = &me->lastUpstreamEvt;
    if (shortUpEvt->sensor == 0) {
        if (me->device.locked) {
            /* device is locked -> send general ack but do nothing */
            SelectA_SendGeneralAck();
            return;
        }
        if (Fid_IsdSLinkDevice(me->device.fid)) {
            if (shortUpEvt->key == 11) {
                /* UDI-Request */
                DsmApi_SendUDIInterruptEvent(shortUpEvt->source, (shortUpEvt->click == 1) ? TRUE : FALSE);
            }
        } else {
            if (!SelectA_HandleResendFlag(me)) {
                me->source = shortUpEvt->source;
                me->key = shortUpEvt->key;
                me->click = shortUpEvt->click;
                me->actionMode = shortUpEvt->actionMode;
                me->actionId = shortUpEvt->actionId;
                /* get group */
                me->groupId = SelectA_GetGroupId(&me->device, me->key);
                SelectA_HandleShortUpstreamOperation(me);
            }
        }
    } else {
        if (!SelectA_HandleResendFlag(me)) {
            if ((shortUpEvt->eventType == Sensor_SensorEvent) ||
                (shortUpEvt->eventType == Sensor_StatusEvent)) {
                if (!me->device.locked) {
                    /* send event only if device is not locked */
                    uint8_t eventIndex = (shortUpEvt->eventType == Sensor_StatusEvent) ? 0x10 : 0;
                    eventIndex += shortUpEvt->data;
                    DsmApi_SendSensorEvent(shortUpEvt->source, eventIndex);
                }
                /* ack event by selecting the device */
                SelectDeviceEvt * reqEvt =
                        Q_NEW(SelectDeviceEvt, SELECT_B_SELECT_SIG);
                reqEvt->deviceId = shortUpEvt->source;
                QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
            } else if ((shortUpEvt->eventType == Sensor_SensorPollRequest) ||
                       (shortUpEvt->eventType == Sensor_StatusPollRequest)) {
                if (me->device.locked) {
                    /* device is locked -> send general ack but do nothing */
                    SelectA_SendGeneralAck();
                    return;
                }
                /* poll sensor value */
                uint8_t sensorIndex = shortUpEvt->data;
                /* get device sensor */
                GetDeviceSensorEvt * reqEvt =
                    Q_NEW(GetDeviceSensorEvt, SELECT_B_GET_DEVICE_SENSOR_SIG);
                reqEvt->deviceId = shortUpEvt->source;
                if (shortUpEvt->eventType == Sensor_StatusPollRequest) {
                    reqEvt->sensorId = 0x10;
                } else {
                    reqEvt->sensorId = 0;
                }
                reqEvt->sensorId += sensorIndex;
                QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
            } else if (shortUpEvt->eventType == Sensor_BinaryInputEvent) {
                if (!me->device.locked) {
                    /* count/send binary input event only if device is not locked */
                    ++me->binaryInputEventCount;

                    uint8_t binaryInputIndex = shortUpEvt->index;
                    uint8_t binaryInputState = shortUpEvt->data;

                    uint8_t binaryInputType;
                    if (Datamodel_GetDeviceBinaryInputType(me->device.deviceId, binaryInputIndex, &binaryInputType)) {
                        DsmApi_SendBinaryInputEvent(shortUpEvt->source, binaryInputIndex, binaryInputType, binaryInputState);
                    } else {
                        DsmApi_SendBinaryInputEvent(shortUpEvt->source, binaryInputIndex, 0, binaryInputState);
                    }
                }
                /* ack event by selecting the device */
                SelectDeviceEvt * reqEvt =
                        Q_NEW(SelectDeviceEvt, SELECT_B_SELECT_SIG);
                reqEvt->deviceId = shortUpEvt->source;
                QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
            }
        }
    }
}

/*..........................................................................*/
void SelectA_ctor(void)
{
    uint8_t i;
    SelectA *me = &l_selectA;
    /* call active object (superclass) constructor */
    QActive_ctor(&me->super, (QStateHandler)&SelectA_initial);
    /* initialize deferred queue */
    QEQueue_init(&me->deferredQueue, l_deferredQueueSto, Q_DIM(l_deferredQueueSto));
    /* initialize shared time events */
    for (i = 0; i < Q_DIM(me->sharedTimeEvts1); ++i) {
        QTimeEvt_ctor(&(me->sharedTimeEvts1[i].super), SELECT_A_SHARED_TIMER_1_TIMEOUT_SIG);
        me->sharedTimeEvts1[i].groupId = 0;
        me->sharedTimeEvts1[i].localZoneId = 0;
        me->sharedTimeEvts1[i].buttonId = 0;
    }
    for (i = 0; i < Q_DIM(me->sharedTimeEvts2); ++i) {
        QTimeEvt_ctor(&(me->sharedTimeEvts2[i].super), SELECT_A_SHARED_TIMER_2_TIMEOUT_SIG);
        me->sharedTimeEvts2[i].groupId = 0;
        me->sharedTimeEvts2[i].localZoneId = 0;
        me->sharedTimeEvts2[i].buttonId = 0;
    }
    /* initialize click time events */
    for (i = 0; i < Q_DIM(me->clickTimeEvts); ++i) {
        QTimeEvt_ctor(&(me->clickTimeEvts[i].super), SELECT_A_CLICK_TIMER_TIMEOUT_SIG);
        me->clickTimeEvts[i].deviceId = 0;
    }
    /* initialize STOP time events */
    for (i = 0; i < Q_DIM(me->stopTimeEvts); ++i) {
        QTimeEvt_ctor(&(me->stopTimeEvts[i].super), SELECT_A_STOP_TIMER_TIMEOUT_SIG);
        me->stopTimeEvts[i].groupId = 0;
        me->stopTimeEvts[i].localZoneId = 0;
        me->stopTimeEvts[i].buttonId = 0;
    }
    /* initialize zone IDs */
    for (i = 0; i < Q_DIM(me->zoneIds); ++i) {
        me->zoneIds[i] = DATAMODEL_INVALID_ZONE_ID;
    }
    /* initialize extended state variables */
    me->apartmentId = 0;
    me->ds485Address = DS485_INVALID_ADDRESS;
    me->apartmentState = APARTMENT_STATE_UNKNOWN;
    me->alarmScene = INVALID_SCENE;
    me->progMode.dSMAddress = DS485_INVALID_ADDRESS;
    QTimeEvt_ctor(&me->progMode.exitTimer, SELECT_A_EXIT_PROG_MODE_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->progMode.aliveTimer, SELECT_A_REMOTE_PROG_MODE_ALIVE_SIG);
    me->binaryInputEventCount = 0;
    /**
     * initialize shared light extended state and state machine
     **/
    me->light.parent = (QActive *)me;
    me->light.progMode = &me->progMode;
    Light_ctor(&(me->light));
    /**
     * initialize shared security extended state and state machine
     **/
    me->security.parent = (QActive *)me;
    Security_ctor(&(me->security));
    /**
     * initialize shared access extended state and state machine
     **/
    me->access.parent = (QActive *)me;
    Access_ctor(&(me->access));
    /**
     * initialize shared group extended state and state machine
     **/
    me->group.parent = (QActive *)me;
    me->group.progMode = &me->progMode;
    Group_ctor(&(me->group));
    /* initialize registered local priority */
    memset(&me->registerLocalPrios, 0, sizeof(me->registerLocalPrios));
    /**
     * Initialize heating
     */
    Heating_ctor((QActive *)me);
}

/*..........................................................................*/
uint8_t SelectA_GetDeferredQueueMargin(void)
{
    return QEQueue_getNFree(&l_selectA.deferredQueue);
}

/*..........................................................................*/
static QState SelectA_initial(SelectA *me, QEvent const *e)
{
    StateHandlerOffset state;
    uint8_t h, i, j;
    /* QS dictionary entries */
    QS_SIG_DICTIONARY(CT_TIP_1X_SIG, 0);
    QS_SIG_DICTIONARY(CT_TIP_2X_SIG, 0);
    QS_SIG_DICTIONARY(CT_TIP_3X_SIG, 0);
    QS_SIG_DICTIONARY(CT_CLICK_1X_SIG, 0);
    QS_SIG_DICTIONARY(CT_CLICK_2X_SIG, 0);
    QS_SIG_DICTIONARY(CT_SHORT_LONG_SIG, 0);
    QS_SIG_DICTIONARY(CT_LOCAL_OFF_SIG, 0);
    QS_SIG_DICTIONARY(CT_LOCAL_ON_SIG, 0);
    QS_SIG_DICTIONARY(CT_INVALID_SIG, 0);
    QS_SIG_DICTIONARY(CALL_SCENE_ZONE_GROUP_SIG, 0);

    /* subscribe events */
    QActive_subscribe((QActive *)me, CONFIG_INIT_SIG);
    QActive_subscribe((QActive *)me, DS485_ADDRESS_SIG);
    QActive_subscribe((QActive *)me, SHORT_UPSTREAM_SIG);
    QActive_subscribe((QActive *)me, DATAMODEL_ZONE_ID_SIG);
    QActive_subscribe((QActive *)me, SENSOR_PUSH_VALUE_SIG);
    /**
     * initialize light state machines
     **/
    Light_init(&(me->light));
    state = GetStateHandlerOffsetFromState(me->light.super.state);
    me->lightGroupInitHsmStateOffset = state;
    for (h = 0; h < NUM_STANDARD_GROUPS; ++h) {
        for (i = 0; i < MAX_ZONES; ++i) {
            me->lightHsmStateOffsets[h][i] = state;
        }
    }
    Group_init(&(me->group));
    state = GetStateHandlerOffsetFromState(me->group.super.state);
    for (h = 0; h < NUM_GROUPS_WITH_AREA; ++h) {
        for (i = 0; i < MAX_ZONES; ++i) {
            for (j = 0; j < 4; ++j) {
                me->buttonGroupHsmStateOffsets[h][i][j] = state;
            }
        }
    }
    /* initialize clusters statemachine groupId and hsm states to 0 */
    for (i = 0; i < NUM_USER_GROUP_WITH_HSM; ++i) {
        me->stateMachineGroupId[i] = 0;
        me->clusterHsmStateOffsets[i] = 0;
    }
    /**
     * initialize security state machines
     **/
    Security_init(&(me->security));
    /**
     * initialize access state machines
     **/
    Access_init(&(me->access));
    /**
     * initialize heating
     */
    Heating_init();
    /**
     * initialize sensors
     */
    SelectA_SensorsInit(&me->sensors);
    return Q_TRAN(&SelectA_wait_for_configuration);
}

/*..........................................................................*/
static QState SelectA_top(SelectA *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_ADDRESS_SIG: {
        /* cache device address */
        me->ds485Address = ((DS485AddressEvt const *)e)->deviceAddress;
        return Q_HANDLED();
    }
    case NOTIFY_USER_GROUP_SIG: {
        NotifyUserGroupEvt const * notifyEvt = (NotifyUserGroupEvt const *)e;
        SelectA_UpdateUserGroup(me, notifyEvt->groupId, notifyEvt->stateMachineGroupId);
        return Q_HANDLED();
    }
    case SELECT_A_CLICK_TIMER_TIMEOUT_SIG: {
        /* clear device ID - cast away const */
        ClickTimeEvt * clickTimeEvt = (ClickTimeEvt *)e;
        clickTimeEvt->deviceId = 0;
        return Q_HANDLED();
    }
    case SELECT_A_STOP_TIMER_TIMEOUT_SIG: {
        /* release stop timer */
        SelectA_ReleaseStopTimer((StopTimerEvt *)e);
        return Q_HANDLED();
    }
    case SELECT_A_REQ_ZONE_SCENE_HISTORY_SIG: {
        ZoneSceneHistoryReqEvt const * reqEvt = (ZoneSceneHistoryReqEvt const *)e;
        ZoneSceneHistoryRingbuffer * rb;
        ZoneSceneHistoryResEvt * resEvt;
        uint8_t count, index, idx;
        Q_ASSERT((reqEvt->localZoneId >= MIN_ZONE) && (reqEvt->localZoneId <= MAX_ZONE));
        rb = &(me->zoneSceneHistoryRingbuffer[reqEvt->localZoneId-1]);
        resEvt = Q_NEW(ZoneSceneHistoryResEvt, SELECT_A_RES_ZONE_SCENE_HISTORY_SIG);
        count = rb->count;
        resEvt->count = count;
        index = rb->index;
        idx = 0;
        while (count) {
            resEvt->groupId[idx] = rb->groupId[index];
            resEvt->sceneId[idx] = rb->scene[index];
            if (index > 0) {
                --index;
            } else {
                index = MAX_ZONE_SCENE_HISTORY - 1;
            }
            ++idx;
            --count;
        }
        QActive_postFIFO(reqEvt->requester, (QEvent *)resEvt);
        return Q_HANDLED();
    }
    case DATAMODEL_ZONE_ID_SIG: {
        DatamodelZoneIdEvt const * evt = (DatamodelZoneIdEvt const *)e;
        bool isZoneNew = (me->zoneIds[evt->localZoneId] == DATAMODEL_INVALID_ZONE_ID);
        /* save zone ID */
        me->zoneIds[evt->localZoneId] = evt->zoneId;
        if (isZoneNew) {
            /* initialize sensors if zone is new */
            SelectA_SensorsInitZone(&me->sensors, evt->localZoneId);
            /* initialize heating if zone is new */
            HeatingReqBasicEvt reqEvt = {{HEATING_INITIALIZE_SIG, 0}, evt->localZoneId};
            Heating_dispatch((QEvent *)&reqEvt, me->zoneIds);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState SelectA_wait_for_configuration(SelectA *me, QEvent const *e)
{
    switch (e->sig) {
    case CONFIG_INIT_SIG: {
        /* cache apartment ID */
        DatamodelConfigEvt const * configInitEvt = (DatamodelConfigEvt const *)e;
        me->apartmentId = configInitEvt->apartmentId;
        return Q_TRAN(&SelectA_idle);
    }
    }
    return Q_SUPER(&SelectA_top);
}

/*..........................................................................*/
/* For the defer/recall mechanism to work correctly Q_HANDLED() must never be called
 * in this state (idle) except for Q_ENTRY_SIG and Q_EXIT_SIG. Always use
 * Q_TRAN(&SelectA_idle) instead.
 */
static QState SelectA_idle(SelectA *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QActive_recall((QActive *)me, &me->deferredQueue);
        return Q_HANDLED();
    }
    case SHORT_UPSTREAM_SIG: {
        ShortUpstreamEvt *shortUpEvt = (ShortUpstreamEvt *)e;
        if (shortUpEvt->source != (MAX_DEVICE+1)) {
            shortUpEvt->circuit = calculateCircuitId(me->ds485Address, me->apartmentId);
            DsmApi_SendShortUpstreamEvent(shortUpEvt, FALSE);
            me->lastUpstreamEvt = *shortUpEvt;
            me->source = shortUpEvt->source;
            /* is device already cached? */
            if ((me->device.magic != DATAMODEL_MAGIC) || (me->device.deviceId != me->source)) {
                /* retrieve device before handling short upstream telegram */
                return Q_TRAN(&SelectA_retrieve_device_for_short_upstream);
            }
            SelectA_HandleShortUpstream(me);
        } else {
            /* hello telegram */
            DsmApi_SendShortUpstreamEvent(shortUpEvt, FALSE);
            if ((shortUpEvt->circuit == calculateCircuitId(me->ds485Address, me->apartmentId)) || (me->ds485Address == DS485_INVALID_ADDRESS)) {
                /* OK, upstream telegram originating from correct circuit: */
                /* start registration */
                static const QEvent helloEvt = { SELECT_B_HELLO_SIG, 0 };
                QActive_postFIFO(AO_SelectB, &helloEvt);
            } else {
                /* upstream telegram circuit is invalid: */
                /* re-send set dSM ID event to devices */
                static const QEvent setDsmIdEvt = { SELECT_B_SET_DSM_ID_SIG, 0 };
                QActive_postFIFO(AO_SelectB, &setDsmIdEvt);
            }
        }
        return Q_TRAN(&SelectA_idle);
    }
    case INTERNAL_CLICK_FORWARD_SIG: {
        ClickForwardEvt const * clickForwardEvt = (ClickForwardEvt const *)e;
        QEvent newE;
        Operation op;

        op.buttonId = clickForwardEvt->buttonId;
        op.groupId = clickForwardEvt->groupId;
        op.localZoneId = clickForwardEvt->localZoneId;
        op.key = clickForwardEvt->key;
        op.fid = clickForwardEvt->fid;
        op.deviceId = clickForwardEvt->deviceId;
        op.apartmentWide = clickForwardEvt->apartmentWide;
        op.outputChannel = clickForwardEvt->outputChannel;
        newE.dynamic_ = 0;
        newE.sig = clickForwardEvt->event;
        SelectA_HandleOperation(me, &op, &newE);
        return Q_TRAN(&SelectA_idle);
    }
    case BUTTON_GROUP_FORCE_SCENE_0_SIG:
    case BUTTON_GROUP_STANDBY_SIG:
    case BUTTON_GROUP_FORCE_SCENE_0_AND_STANDBY_SIG:
    case BUTTON_GROUP_DEEP_OFF_SIG:
    case BUTTON_GROUP_LOCAL_PRIO_OFF_SIG:
        /* fall through */
    case BUTTON_GROUP_PROGRAMM_MODE_SIG: {
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        Operation op;
        op.deviceId = sceneBasicEvt->sourceDeviceId;
        op.buttonId = BUTTON_ZONE_0;
        op.groupId = sceneBasicEvt->sourceGroupId;
        op.localZoneId = sceneBasicEvt->sourceZoneId;
        op.apartmentWide = FALSE;
        op.outputChannel = sceneBasicEvt->outputChannel;
        SelectA_HandleOperation(me, &op, e);
        return Q_TRAN(&SelectA_idle);
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG:
        /* fall through */
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle operation only if not in programming mode */
        if (me->progMode.localZoneId == 0) {
            SceneZoneGroupEvt const * sceneZoneGroupEvt = (SceneZoneGroupEvt const *)e;
            if (sceneZoneGroupEvt->localZoneId == 0) {
                if (sceneZoneGroupEvt->groupId == 0) {
                    SelectA_HandleSceneAllOperation(me, sceneZoneGroupEvt);
                } else {
                    SelectA_HandleSceneGroupOperation(me, sceneZoneGroupEvt);
                }
            } else {
                SelectA_HandleSceneZoneOperation(me, sceneZoneGroupEvt);
            }
            if ((e->sig == UNDO_SCENE_NUMBER_BROADCAST_SIG) && (me->alarmScene == sceneZoneGroupEvt->basic.sceneId)) {
                /* reset alarm scene */
                me->alarmScene = INVALID_SCENE;
            }
        }
        return Q_TRAN(&SelectA_idle);
    }
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG:
        /* fall through */
    case CALL_SCENE_ZONE_GROUP_SIG: {
        /* handle operation only if not in programming mode */
        if (me->progMode.localZoneId == 0) {
            SelectA_HandleSceneZoneGroupOperation(me, (SceneZoneGroupEvt const *)e);
            Heating_dispatch(e, me->zoneIds);
        }
        return Q_TRAN(&SelectA_idle);
    }
    case UNDO_SCENE_NUMBER_DEVICE_SIG:
        /* fall through */
    case CALL_SCENE_DEVICE_SIG: {
        SelectA_HandleSceneDeviceOperation(me, (SceneDeviceEvt const *)e);
        return Q_TRAN(&SelectA_idle);
    }
    case SELECT_A_SHARED_TIMER_1_TIMEOUT_SIG: {
        SelectA_HandleSharedTimer1Operation(me, (SharedOpTimeEvt const *)e);
        return Q_TRAN(&SelectA_idle);
    }
    case SELECT_A_SHARED_TIMER_2_TIMEOUT_SIG: {
        SelectA_HandleSharedTimer2Operation(me, (SharedOpTimeEvt const *)e);
        return Q_TRAN(&SelectA_idle);
    }
    case SELECT_A_EXIT_PROG_MODE_TIMEOUT_SIG:
        /* fall through */
    case SELECT_A_REMOTE_PROG_MODE_ALIVE_SIG:
        /* fall through */
    case DATAMODEL_RES_ZONE_DEVICE_COUNT_SIG:
        /* fall through */
    case DS485_RES_GET_BUS_MEMBER_BY_INDEX_SIG:
        /* fall through */
    case DS485_RX_DATA_SIG:
        /* fall through */
    case DSMAPI_REQUEST_TIMEOUT_SIG: {
        SelectA_HandleProgrammingModeSignal(me, e);
        return Q_TRAN(&SelectA_idle);
    }
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        DatamodelResDeviceEntryEvt const * resEvt = (DatamodelResDeviceEntryEvt const *)e;
        Q_ASSERT(resEvt->privateId == SELECT_A_PRIVATE_ID_PROGRAMMING_MODE);
        SelectA_HandleDatamodelDeviceEntryResponse(me, resEvt);
        return Q_TRAN(&SelectA_idle);
    }
    case SAVE_SCENE_ZONE_SIG: {
        ActionZoneGroupSceneEvt const * reqEvt = (ActionZoneGroupSceneEvt const *)e;
        Scene scene = (Scene)reqEvt->sceneId;
        SelectASocket_SendSaveSceneZoneGroup(reqEvt->localZoneId, reqEvt->groupId, scene);
        return Q_TRAN(&SelectA_idle);
    }
    case SAVE_SCENE_DEVICE_SIG: {
        SaveSceneDeviceEvt const * reqEvt = (SaveSceneDeviceEvt const *)e;
        Scene scene = (Scene)reqEvt->sceneId;
        SelectASocket_SendSaveSceneDevice(reqEvt->deviceId, scene);
        return Q_TRAN(&SelectA_idle);
    }
    case UNDO_SCENE_ZONE_SIG: {
        ActionZoneEvt const * reqEvt = (ActionZoneEvt const *)e;
        SelectASocket_SendUndoSceneZoneGroup(reqEvt->localZoneId, reqEvt->groupId);
        return Q_TRAN(&SelectA_idle);
    }
    case BLINK_ZONE_SIG: {
        ActionZoneEvt const * reqEvt = (ActionZoneEvt const *)e;
        SelectASocket_SendBlinkZoneGroup(reqEvt->localZoneId, reqEvt->groupId);
        return Q_TRAN(&SelectA_idle);
    }
    case PROG_ON_ZONE_SIG:
        /* fall through */
    case PROG_OFF_ZONE_SIG: {
        ProgZoneEvt const * reqEvt = (ProgZoneEvt const *)e;
        uint8_t localZoneId = SelectA_GetLocalZoneId(me, reqEvt->zoneId);
        if (localZoneId != 0) {
            /* valid zone found */
            Operation op;
            op.localZoneId = localZoneId;
            op.groupId = reqEvt->groupId;
            op.fid = 0;
            op.key = 0xff;
            op.buttonId = BUTTON_ZONE_0;
            op.deviceId = 0;
            op.apartmentWide = FALSE;
            op.outputChannel = 0;
            SelectA_HandleOperation(me, &op, (QEvent const *)e);
        } else {
            /* no valid zone found */
            uint16_t zoneId = reqEvt->zoneId;
            if (e->sig == PROG_ON_ZONE_SIG) {
                /* PROG ON received */
                if (me->progMode.localZoneId == 0) {
                    /* enter programming mode */
                    me->progMode.localZoneId = (MAX_ZONE + 1);
                    me->progMode.groupId = reqEvt->groupId;
                    /* save zone ID */
                    me->zoneIds[MAX_ZONE + 1] = zoneId;
                    /* start timer */
                    QTimeEvt_postIn(&(me->progMode.exitTimer), (QActive *)me, REMOTE_PROG_EXIT_TIMEOUT_TICKS);
                } else {
                    /* restart timer */
                    QTimeEvt_rearm(&(me->progMode.exitTimer), REMOTE_PROG_EXIT_TIMEOUT_TICKS);
                }
            } else {
                /* PROG OFF received */
                if ((me->progMode.localZoneId == (MAX_ZONE + 1)) && (me->zoneIds[MAX_ZONE + 1] == zoneId)) {
                    /* exit programming mode */
                    me->zoneIds[MAX_ZONE + 1] = 0;
                    me->progMode.localZoneId = 0;
                    /* stop exit timer */
                    QTimeEvt_disarm(&(me->progMode.exitTimer));
                    return Q_TRAN(&SelectA_idle);
                }
            }
        }
        return Q_TRAN(&SelectA_idle);
    }
    case SELECT_ZONE_GROUP_SIG: {
        return Q_TRAN(&SelectA_idle);
    }
    case SET_OUTVAL_ZONE_GROUP_SIG: {
        SetOutvalZoneGroupEvt const * reqEvt = (SetOutvalZoneGroupEvt const *)e;
        SelectASocket_SendSetOutvalZoneGroup(reqEvt->localZoneId, reqEvt->groupId, reqEvt->outval);
        return Q_TRAN(&SelectA_idle);
    }
    case UNDO_SCENE_DEVICE_SIG: {
        ActionDeviceEvt const * reqEvt = (ActionDeviceEvt const *)e;
        SelectASocket_SendUndoSceneDevice(reqEvt->deviceId);
        return Q_TRAN(&SelectA_idle);
    }
    case BLINK_DEVICE_SIG: {
        ActionDeviceEvt const * reqEvt = (ActionDeviceEvt const *)e;
        SelectASocket_SendBlinkDevice(reqEvt->deviceId);
        return Q_TRAN(&SelectA_idle);
    }
    case PROG_ON_DEVICE_SIG: {
        ActionDeviceEvt const * reqEvt = (ActionDeviceEvt const *)e;
        SelectA_SendProgOnDevice(reqEvt->deviceId);
        return Q_TRAN(&SelectA_idle);
    }
    case PROG_OFF_DEVICE_SIG: {
        ActionDeviceEvt const * reqEvt = (ActionDeviceEvt const *)e;
        SelectA_SendProgOffDevice(reqEvt->deviceId);
        return Q_TRAN(&SelectA_idle);
    }
    case SET_OUTVAL_DEVICE_SIG: {
        SetOutvalDeviceEvt const * reqEvt = (SetOutvalDeviceEvt const *)e;
        SelectASocket_SendSetOutvalDevice(reqEvt->deviceId, reqEvt->outval);
        return Q_TRAN(&SelectA_idle);
    }
    case ADD_DEVICE_TO_GROUP_SIG: {
        AddDeviceToGroupEvt const * addDeviceToGroupEvt = (AddDeviceToGroupEvt const *)e;
        SelectASocket_SendAddGroupDevice(addDeviceToGroupEvt->deviceId, addDeviceToGroupEvt->groupId);
        return Q_TRAN(&SelectA_idle);
    }
    case REM_DEVICE_FROM_GROUP_SIG: {
        RemoveDeviceFromGroupEvt const * remDeviceFromGroupEvt = (RemoveDeviceFromGroupEvt const *)e;
        SelectASocket_SendRemoveGroupDevice(remDeviceFromGroupEvt->deviceId, remDeviceFromGroupEvt->groupId);
        return Q_TRAN(&SelectA_idle);
    }
    case REM_GROUP_FROM_ZONE_SIG: {
        RemoveGroupFromZoneEvt const * remGroupFromZoneEvt = (RemoveGroupFromZoneEvt const *)e;
        SelectASocket_SendRemoveGroupFromZone(remGroupFromZoneEvt->localZoneId, remGroupFromZoneEvt->groupId);
        return Q_TRAN(&SelectA_idle);
    }
    case OPC_DIM_ZONE_GROUP_SIG: {
        OPCDimZoneGroupEvt const * reqEvt = (OPCDimZoneGroupEvt const *)e;
        SelectA_SendOPCDimZoneGroup(reqEvt->localZoneId, reqEvt->groupId, reqEvt->operation, reqEvt->channelId);
        return Q_TRAN(&SelectA_idle);
    }
    case OPC_DIM_DEVICE_SIG: {
        OPCDimDeviceEvt const * reqEvt = (OPCDimDeviceEvt const *)e;
        SelectASocket_SendOPCDimDevice(reqEvt->deviceId, reqEvt->operation, reqEvt->channelId);
        return Q_TRAN(&SelectA_idle);
    }
    case DATAMODEL_CHANGED_SIG: {
        SelectA_InvalidateCaches(me);
        return Q_TRAN(&SelectA_idle);
    }
    case SET_SCENE_LOCAL_PRIO_BROADCAST_SIG: {
        /* handle operation only if not in programming mode */
        if (me->progMode.localZoneId == 0) {
            ActionZoneGroupSceneEvt const * setSceneLocalPrioEvt = (ActionZoneGroupSceneEvt const *)e;
            if (setSceneLocalPrioEvt->localZoneId == 0) {
                if (setSceneLocalPrioEvt->groupId == 0) {
                    SelectA_HandleSetLocalPrioAllOperation(me, setSceneLocalPrioEvt);
                } else {
                    SelectA_HandleSetLocalPrioGroupOperation(me, setSceneLocalPrioEvt);
                }
            } else {
                if (setSceneLocalPrioEvt->groupId == 0) {
                    SelectA_HandleSetLocalPrioZoneOperation(me, setSceneLocalPrioEvt);
                } else {
                    SelectA_HandleSetLocalPrioZoneGroupOperation(me, setSceneLocalPrioEvt);
                }
            }
        }
        return Q_TRAN(&SelectA_idle);
    }
    case CALL_SCENE_MIN_BROADCAST_SIG: {
        /* handle operation only if not in programming mode */
        if (me->progMode.localZoneId == 0) {
            ActionZoneGroupSceneEvt const * setSceneLocalPrioEvt = (ActionZoneGroupSceneEvt const *)e;
            SelectA_HandleCallSceneMinAllOperation(me, setSceneLocalPrioEvt);
        }
        return Q_TRAN(&SelectA_idle);
    }
    case REGISTER_SCENE_LOCAL_PRIO_BROADCAST_SIG: {
        if (me->progMode.localZoneId == 0) {
            ActionZoneGroupSceneEvt const * setSceneLocalPrioEvt = (ActionZoneGroupSceneEvt const *)e;
            if (setSceneLocalPrioEvt->localZoneId != 0) {
                if (setSceneLocalPrioEvt->groupId == 0) {
                    SelectA_HandleRegisterLocalPrioZoneOperation(me, setSceneLocalPrioEvt);
                } else {
                    SelectA_HandleRegisterLocalPrioZoneGroupOperation(me, setSceneLocalPrioEvt);
                }
            }
        }
        return Q_TRAN(&SelectA_idle);
    }
    case SELECT_A_DEVICE_PAIRING_SIG: {
        DevicePairingEvt const * pairingEvt = (DevicePairingEvt const *)e;
        me->pairingDeviceId = pairingEvt->deviceId;
        me->pairingZoneId = pairingEvt->zoneId;
        me->pairingZoneCreated = FALSE;
        return Q_TRAN(&SelectA_device_set_zone);
    }
    case LOCAL_STOP_SIG: {
        /* handle operation only if in programming mode */
        if (me->progMode.localZoneId != 0) {
            ActionZoneEvt const * reqEvt = (ActionZoneEvt const *)e;
            QEvent newE;
            Operation op;

            op.buttonId = 0;
            op.groupId = reqEvt->groupId;
            op.localZoneId = reqEvt->localZoneId;
            op.key = 0;
            op.fid = 0;
            op.deviceId = 0;
            op.apartmentWide = FALSE;
            op.outputChannel = 0;
            newE.dynamic_ = 0;
            newE.sig = reqEvt->super.sig;
            SelectA_HandleOperation(me, &op, &newE);
            return Q_TRAN(&SelectA_idle);
        }
        return Q_TRAN(&SelectA_idle);
    }
    case SELECT_A_REQ_ROOM_STATES_SIG: {
        uint8_t groupId, groupIndex, groupShiftIndex;
        ZoneStatesReqEvt const * reqEvt = (ZoneStatesReqEvt const *)e;
        ZoneStatesResEvt * resEvt = Q_NEW(ZoneStatesResEvt, SELECT_A_RES_ROOM_STATES_SIG);
        resEvt->roomStates = 0;
        resEvt->area1States = 0;
        resEvt->area2States = 0;
        resEvt->area3States = 0;
        resEvt->area4States = 0;
        for (groupId = 0; groupId < NUM_GROUPS_WITH_AREA; ++groupId) {
            if (groupId < JOKER_GROUP) {
                groupIndex = groupId;
                groupShiftIndex = groupId;
            } else {
                groupIndex = groupId - 1;
                groupShiftIndex = groupId + 1;
            }
            resEvt->roomStates |= (Light_is_off_state(me->lightHsmStateOffsets[groupIndex][reqEvt->localZoneId - 1]) ? 0 : 1 << groupShiftIndex);
            resEvt->area1States |= (Group_is_off_state(me->buttonGroupHsmStateOffsets[groupIndex][reqEvt->localZoneId - 1][0]) ? 0 : 1 << groupShiftIndex);
            resEvt->area2States |= (Group_is_off_state(me->buttonGroupHsmStateOffsets[groupIndex][reqEvt->localZoneId - 1][1]) ? 0 : 1 << groupShiftIndex);
            resEvt->area3States |= (Group_is_off_state(me->buttonGroupHsmStateOffsets[groupIndex][reqEvt->localZoneId - 1][2]) ? 0 : 1 << groupShiftIndex);
            resEvt->area4States |= (Group_is_off_state(me->buttonGroupHsmStateOffsets[groupIndex][reqEvt->localZoneId - 1][3]) ? 0 : 1 << groupShiftIndex);
        }
        QActive_postFIFO(AO_DsmApi, (QEvent *)resEvt);
        return Q_TRAN(&SelectA_idle);
    }
    case SENSOR_PUSH_VALUE_SIG: {
        ZoneGroupSensorPushEvt const *reqEvt = (ZoneGroupSensorPushEvt const *)e;
        /* update sensor value */
        SelectA_SensorsUpdateSensor(reqEvt);
        Heating_dispatch(e, me->zoneIds);
        return Q_TRAN(&SelectA_idle);
    }
    case HEATING_FOLLOW_SENSOR_PUSH_SIG:
    case HEATING_REQ_SET_STATE_SIG:
    case HEATING_NOTIFY_CONFIG_UPDATE_SIG:
    case HEATING_NOTIFY_CONFIG_CHANGE_SIG:
    case HEATING_MINUTE_TICK_SIG:
    case HEATING_NOTIFY_SCENE_CHANGE_SIG:
        /* fall through */
    case HEATING_ZONE_MINUTE_TICK_SIG: {
        Heating_dispatch(e, me->zoneIds);
        return Q_TRAN(&SelectA_idle);
    }
    }
    return Q_SUPER(&SelectA_top);
}

/*..........................................................................*/
static QState SelectA_busy(SelectA *me, QEvent const *e)
{
    switch (e->sig) {
    default: {
        /* defer all user events */
        if (e->sig >= Q_USER_SIG) {
            QActive_defer((QActive *)me, &me->deferredQueue, e);
            return Q_HANDLED();
        }
        break;
    }
    }
    return Q_SUPER(&SelectA_top);
}

/*..........................................................................*/
static QState SelectA_retrieve_device_for_short_upstream(SelectA *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DatamodelReqDeviceIdEvt * reqEvt = Q_NEW(DatamodelReqDeviceIdEvt, DATAMODEL_REQ_DEVICE_ENTRY_BY_ID_SIG);
        reqEvt->basic.requester = (QActive *)me;
        reqEvt->deviceId = me->source;
        reqEvt->privateId = SELECT_A_PRIVATE_ID_UPSTREAM;
        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
        return Q_HANDLED();
    }
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        DatamodelResDeviceEntryEvt const * resEvt = (DatamodelResDeviceEntryEvt const *)e;
        if (resEvt->privateId == SELECT_A_PRIVATE_ID_UPSTREAM) {
            if (!resEvt->state) {
                PingInactiveDeviceEvt * pingEvt = Q_NEW(PingInactiveDeviceEvt, SELECT_B_PING_INACTIVE_DEVICE_SIG);
                pingEvt->deviceId = resEvt->device.deviceId;
                QActive_postFIFO(AO_SelectB, (QEvent *)pingEvt);
            }
            memcpy(&me->device, &resEvt->device, sizeof(me->device));
            /* get group ID */
            me->groupId = SelectA_GetGroupId(&me->device, me->key);
            SelectA_HandleShortUpstream(me);
            return Q_TRAN(&SelectA_idle);
        }
        break;
    }
    case DATAMODEL_RES_ERROR_SIG: {
        DatamodelLongErrorEvt const * resEvt = (DatamodelLongErrorEvt const *)e;
        if (resEvt->privateId == SELECT_A_PRIVATE_ID_UPSTREAM) {
            /* device not found -> don't do anything */
            return Q_TRAN(&SelectA_idle);
        }
        break;
    }
    }
    return Q_SUPER(&SelectA_busy);
}

/*..........................................................................*/
static QState SelectA_device_set_zone(SelectA *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DatamodelReqDeviceSetZoneEvt * reqEvt =
            Q_NEW(DatamodelReqDeviceSetZoneEvt, DATAMODEL_REQ_DEVICE_SET_ZONE_SIG);
        reqEvt->basic.requester = (QActive *)me;
        reqEvt->deviceAddress = me->pairingDeviceId;
        reqEvt->zoneId = me->pairingZoneId;
        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
    }
    case DATAMODEL_RES_ERROR_SIG: {
        switch (((DatamodelErrorEvt const *)e)->error) {
        case ERROR_OK: {
            /* pairing successfully done */
            SelectASocket_SendBlinkDevice(me->pairingDeviceId);
            /* send device changed event */
            DsmApi_SendDeviceModelChangedEvent(me->pairingDeviceId);
            /* has zone been created? */
            if (me->pairingZoneCreated) {
                /* move programming mode control from select A to light state machine! */
                if (me->progMode.localZoneId == (MAX_ZONE + 1)) {
                    ProgZoneEvt * progZoneEvt;
                    /* stop exit timer */
                    QTimeEvt_disarm(&(me->progMode.exitTimer));
                    /* allocate PROG ON event */
                    progZoneEvt = Q_NEW(ProgZoneEvt, PROG_ON_ZONE_SIG);
                    progZoneEvt->zoneId = me->pairingZoneId;
                    progZoneEvt->groupId = me->progMode.groupId;
                    /* post event */
                    QActive_postLIFO((QActive *)me, (QEvent *)progZoneEvt);
                    /* exit programming mode - will be set when executing the previous event */
                    me->progMode.localZoneId = 0;
                }
            }
            break;
        }
        case ERROR_ZONE_NOT_FOUND: {
            /* create zone */
            return Q_TRAN(&SelectA_device_create_zone);
        }
        default: {
            break;
        }
        }
        return Q_TRAN(&SelectA_idle);
    }
    }
    return Q_SUPER(&SelectA_busy);
}

/*..........................................................................*/
static QState SelectA_device_create_zone(SelectA *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DatamodelReqZoneModifyEvt * reqEvt =
            Q_NEW(DatamodelReqZoneModifyEvt, DATAMODEL_REQ_ZONE_MODIFY_SIG);
        reqEvt->basic.requester = (QActive *)me;
        reqEvt->modifier = ZONE_MODIFY_ADD;
        reqEvt->zoneId = me->pairingZoneId;
        reqEvt->newZoneId = 0;

        char destAddr [DSUID_STR_LEN];
        dsuid_t vdc_dsuid;
        Datamodel_GetDsuidFromDeviceId(me->pairingDeviceId, destAddr);
        dsuid_from_string(destAddr, &vdc_dsuid);
        reqEvt->dsuid = vdc_dsuid;

        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
        return Q_HANDLED();
    }
    case DATAMODEL_RES_ERROR_SIG: {
        if (((DatamodelErrorEvt const *)e)->error == ERROR_OK) {
            /* zone successfully created */
            me->pairingZoneCreated = TRUE;
            return Q_TRAN(&SelectA_device_set_zone);
        }
        return Q_TRAN(&SelectA_idle);
    }
    }
    return Q_SUPER(&SelectA_busy);
}


/*..........................................................................*/
static SharedOpTimeEvt * SelectA_GetSharedOpTimeEvt1(SelectA *me, uint8_t localZoneId, uint8_t groupId, uint8_t buttonId, uint8_t outputChannel)
{
    uint8_t i;
    SharedOpTimeEvt * freeOpTimeEvt = 0;
    /* find shared timer which is still in use */
    for (i = 0; i < Q_DIM(me->sharedTimeEvts1); ++i) {
        SharedOpTimeEvt * opTimeEvt = &me->sharedTimeEvts1[i];
        if ((opTimeEvt->localZoneId == localZoneId) &&
            (opTimeEvt->groupId == groupId) &&
            (opTimeEvt->buttonId == Datamodel_ButtonId_ButtonGroupId(buttonId)) &&
            (opTimeEvt->outputChannel == outputChannel))
        {
            return opTimeEvt;
        }
        /* save first free shared timer */
        if (freeOpTimeEvt == 0) {
            if (opTimeEvt->sig == 0) {
                freeOpTimeEvt = opTimeEvt;
            }
        }
    }
    return freeOpTimeEvt;
}

/*..........................................................................*/
static SharedOpTimeEvt * SelectA_GetSharedOpTimeEvt2(SelectA *me, uint8_t localZoneId, uint8_t groupId, uint8_t buttonId, uint8_t outputChannel)
{
    uint8_t i;
    SharedOpTimeEvt * freeOpTimeEvt = 0;
    /* find shared timer which is still in use */
    for (i = 0; i < Q_DIM(me->sharedTimeEvts2); ++i) {
        SharedOpTimeEvt * opTimeEvt = &me->sharedTimeEvts2[i];
        if ((opTimeEvt->localZoneId == localZoneId) &&
            (opTimeEvt->groupId == groupId) &&
            (opTimeEvt->buttonId == Datamodel_ButtonId_ButtonGroupId(buttonId)) &&
            (opTimeEvt->outputChannel == outputChannel)) {
            return opTimeEvt;
        }
        /* save first free shared timer */
        if (freeOpTimeEvt == 0) {
            if (opTimeEvt->sig == 0) {
                freeOpTimeEvt = opTimeEvt;
            }
        }
    }
    return freeOpTimeEvt;
}

/*..........................................................................*/
StopTimerEvt* SelectA_AllocateStopTimer(uint8_t localZoneId, uint8_t groupId, uint8_t buttonId)
{
    SelectA *me = &l_selectA;
    StopTimerEvt *timer;

    /* find a free timer */
    timer = SelectA_GetFreeStopTimer(me);
    if (timer != NULL) {
        /* register timer */
        timer->localZoneId = localZoneId;
        timer->groupId = groupId;
        timer->buttonId = Datamodel_ButtonId_ButtonGroupId(buttonId);
    }
    return timer;
}

/*..........................................................................*/
void SelectA_ReleaseStopTimer(StopTimerEvt *timer)
{
    /* stop timer */
    QTimeEvt_disarm(&timer->super);
    timer->localZoneId = 0;
    timer->groupId = 0;
    timer->buttonId = 0;
}

/*..........................................................................*/
static StopTimerEvt * SelectA_GetStopTimer(uint8_t localZoneId, uint8_t groupId, uint8_t buttonId)
{
    SelectA *me = &l_selectA;
    uint8_t i;
    for (i = 0; i < Q_DIM(me->stopTimeEvts); ++i) {
        StopTimerEvt *timer = &me->stopTimeEvts[i];
        if ((timer->localZoneId == localZoneId) && (timer->groupId == groupId) && (timer->buttonId == Datamodel_ButtonId_ButtonGroupId(buttonId))) {
            return timer;
        } else if ((groupId >= USER_GROUP) && (groupId < MAX_USER_GROUP_WITH_HSM) && (timer->groupId == groupId)) {
            /* use same timer for standard user groups */
            return timer;
        }
    }
    return NULL;
}

/*..........................................................................*/
static StopTimerEvt * SelectA_GetFreeStopTimer(SelectA *me)
{
    uint8_t i;
    for (i = 0; i < Q_DIM(me->stopTimeEvts); ++i) {
        StopTimerEvt *timer = &me->stopTimeEvts[i];
        if (timer->localZoneId == 0) {
            return timer;
        }
    }
    return NULL;
}

/*..........................................................................*/
StopTimerEvt const * SelectA_GetStopTimerByIndex(uint8_t index)
{
    SelectA *me = &l_selectA;
    if (index < Q_DIM(me->stopTimeEvts)) {
        return &me->stopTimeEvts[index];
    }
    return NULL;
}

/*..........................................................................*/
bool SelectA_ButtonGroupTimerRunning(uint8_t localZoneId, uint8_t groupId)
{
    uint8_t buttonId;
    if (groupId > LAST_GROUP_WITH_AREA) {
        /* there are no areas for non standard groups */
        return FALSE;
    }
    for (buttonId = 1; buttonId <= 4; ++buttonId) {
        StopTimerEvt *timer = SelectA_GetStopTimer(localZoneId, groupId, buttonId);
        if (timer != NULL) {
            return (timer->super.prev != 0);
        }
    }
    return FALSE;
}

/*..........................................................................*/
bool SelectA_RoomTimerRunning(uint8_t localZoneId, uint8_t groupId)
{
    StopTimerEvt *timer = SelectA_GetStopTimer(localZoneId, groupId, BUTTON_ZONE_0);
    if (timer != NULL) {
        return (timer->super.prev != 0);
    }
    return FALSE;
}

/*..........................................................................*/
void SelectA_StopRoomTimer(uint8_t localZoneId, uint8_t groupId)
{
    StopTimerEvt *timer = SelectA_GetStopTimer(localZoneId, groupId, BUTTON_ZONE_0);
    QTimeEvt_disarm(&timer->super);
    timer->localZoneId = 0;
    timer->groupId = 0;
    timer->buttonId = 0;
}

/*..........................................................................*/
void SelectA_InitiateDevicePairing(uint16_t deviceId, uint8_t localZoneId)
{
    SelectA *me = &l_selectA;
    DevicePairingEvt * pairingEvt = Q_NEW(DevicePairingEvt, SELECT_A_DEVICE_PAIRING_SIG);
    pairingEvt->deviceId = deviceId;
    pairingEvt->zoneId = me->zoneIds[localZoneId];
    QActive_postLIFO((QActive *)me, (QEvent *)pairingEvt);
}

/*..........................................................................*/
void SelectA_RegisterLocalPrio(uint8_t localZoneId, uint8_t groupId, uint8_t buttonId, uint16_t deviceId)
{
    SelectA *me = &l_selectA;
    uint8_t i;
    for (i = 0; i < Q_DIM(me->registerLocalPrios); ++i) {
        if (me->registerLocalPrios[i].buttonGroupId == 0) {
            /* free entry found */
            me->registerLocalPrios[i].buttonGroupId = Datamodel_ButtonId_ButtonGroupId(buttonId);
            me->registerLocalPrios[i].localZoneId = localZoneId;
            me->registerLocalPrios[i].groupId = groupId;
            me->registerLocalPrios[i].deviceId = deviceId;
            break;
        }
    }
}

/*..........................................................................*/
void SelectA_UnregisterLocalPrio(uint8_t localZoneId, uint8_t groupId, uint8_t buttonId)
{
    SelectA *me = &l_selectA;
    uint8_t i;
    uint8_t buttonGroupId = Datamodel_ButtonId_ButtonGroupId(buttonId);
    for (i = 0; i < Q_DIM(me->registerLocalPrios); ++i) {
        if (me->registerLocalPrios[i].buttonGroupId == buttonGroupId) {
            if ((me->registerLocalPrios[i].localZoneId == localZoneId) &&
                (me->registerLocalPrios[i].groupId == groupId)) {
                /* matching entry found */
                me->registerLocalPrios[i].buttonGroupId = 0;
                break;
            }
        }
    }
}

/*..........................................................................*/
void SelectA_HandleRegisteredLocalPrio(uint8_t localZoneId, uint8_t groupId, uint16_t deviceId)
{
    SelectA *me = &l_selectA;
    uint8_t i;
    for (i = 0; i < Q_DIM(me->registerLocalPrios); ++i) {
        uint8_t buttonGroupId = me->registerLocalPrios[i].buttonGroupId;
        if (buttonGroupId != 0) {
            if ((groupId == me->registerLocalPrios[i].groupId) && (deviceId != me->registerLocalPrios[i].deviceId)) {
                if ((localZoneId == 0) || (localZoneId == me->registerLocalPrios[i].localZoneId)) {
                    QEvent e;
                    StateHandlerOffset *state;
                    uint8_t groupIndex;
                    localZoneId = me->registerLocalPrios[i].localZoneId;
                    me->group.deviceId = 0;
                    me->group.localZoneId = localZoneId;
                    me->group.groupId = groupId;
                    me->group.buttonId = buttonGroupId;
                    me->group.opTimer1 = SelectA_GetSharedOpTimeEvt1(me, localZoneId, groupId, buttonGroupId, 0);
                    me->group.opTimer2 = SelectA_GetSharedOpTimeEvt2(me, localZoneId, groupId, buttonGroupId, 0);
                    me->group.stopTimer = SelectA_GetStopTimer(localZoneId, groupId, buttonGroupId);
                    me->group.fid = 0;
                    me->group.key = 0;
                    me->group.setLocalPriority = FALSE;
                    groupIndex = groupId - (groupId < JOKER_GROUP ? 1 : 2);
                    state = &me->buttonGroupHsmStateOffsets[groupIndex][localZoneId - MIN_ZONE][buttonGroupId - 1];
                    me->group.super.state = GetStateFromStateHandlerOffset(*state);
                    e.sig = CALL_ROOM_SCENE_SIG;
                    e.dynamic_ = 0;
                    Group_dispatch(&(me->group), &e);
                    *state = GetStateHandlerOffsetFromState(me->group.super.state);
                }
            }
        }
    }
}

/*..........................................................................*/
Scene SelectA_SecurityAlarmScene()
{
    SelectA *me = &l_selectA;
    return me->alarmScene;
}

/*..........................................................................*/
Scene SelectA_LightAlarmScene()
{
    SelectA *me = &l_selectA;
    if ((me->alarmScene == PANIC) || (me->alarmScene == FIRE)) {
        return me->alarmScene;
    }
    return INVALID_SCENE;
}

/*..........................................................................*/
ApartmentState_t SelectA_GetApartmentState()
{
    return l_selectA.apartmentState;
}

/*..........................................................................*/
static void SelectA_UnhandledZoneGroupOperation(QEvent const *e)
{
    switch (e->sig) {
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        /* unhandled undo scene number zone group operation */
        /* send undo scene number to line */
        SceneZoneGroupEvt const * zoneGroupEvt = (SceneZoneGroupEvt const *)e;
        SelectA_SendUndoSceneNumberZoneGroup(zoneGroupEvt->localZoneId, zoneGroupEvt->groupId, (Scene)zoneGroupEvt->basic.sceneId);
        break;
    }
    case CALL_SCENE_ZONE_GROUP_SIG: {
        /* unhandled call scene zone group operation */
        /* send call scene to line */
        SceneZoneGroupEvt const * zoneGroupEvt = (SceneZoneGroupEvt const *)e;
        SelectA_SendSceneZoneGroup(zoneGroupEvt->localZoneId, zoneGroupEvt->groupId, (Scene)zoneGroupEvt->basic.sceneId, zoneGroupEvt->basic.force);
        break;
    }
    default: {
        /* don't do nothing */
        break;
    }
    }
}

/*..........................................................................*/
static void SelectA_HandleOperation(SelectA *me, Operation const * op, QEvent const *e)
{
    uint8_t stateMachineGroupId, zoneIdx, groupIdx;
    StateHandlerOffset * state;
    zoneIdx = op->localZoneId - MIN_ZONE;
    if (op->groupId <= LAST_STANDARD_GROUP) {
        stateMachineGroupId = op->groupId;
        groupIdx = stateMachineGroupId - (op->groupId < JOKER_GROUP ? 1 : 2);
        if ((stateMachineGroupId <= LAST_GROUP_WITH_AREA) && (me->progMode.localZoneId == 0) && (Datamodel_ButtonId_ButtonGroupId(op->buttonId) > 0)) {
            state = &me->buttonGroupHsmStateOffsets[groupIdx][zoneIdx][Datamodel_ButtonId_ButtonGroupId(op->buttonId) - 1];
            me->group.deviceId = op->deviceId;
            me->group.localZoneId = op->localZoneId;
            me->group.groupId = op->groupId;
            me->group.buttonId = op->buttonId;
            me->group.outputChannel = op->outputChannel;
            me->group.opTimer1 = SelectA_GetSharedOpTimeEvt1(me, op->localZoneId, op->groupId, op->buttonId, op->outputChannel);
            me->group.opTimer2 = SelectA_GetSharedOpTimeEvt2(me, op->localZoneId, op->groupId, op->buttonId, op->outputChannel);
            me->group.stopTimer = SelectA_GetStopTimer(op->localZoneId, op->groupId, op->buttonId);
            me->group.fid = op->fid;
            me->group.key = op->key;
            me->group.setLocalPriority = op->setLocalPriority;
            me->group.deviceId = op->deviceId;
            me->group.super.state = GetStateFromStateHandlerOffset(*state);
            Group_dispatch(&(me->group), e);
            *state = GetStateHandlerOffsetFromState(me->group.super.state);
            return;
        } else {
            switch (stateMachineGroupId) {
            case LIGHT_GROUP:
            case SHADE_GROUP:
            case HEATING_GROUP:
            case AUDIO_GROUP:
            case VIDEO_GROUP:
            case SECURITY_GROUP:
            case ACCESS_GROUP:
            case COOLING_GROUP:
            case VENTILATION_GROUP:
            case WINDOW_GROUP: {
                state = &me->lightHsmStateOffsets[groupIdx][zoneIdx];
                break;
            }
            default: {
                SelectA_UnhandledZoneGroupOperation(e);
                return;
            }
            }
        }
    } else {
        if (op->groupId > MAX_GROUP) {
            /* handle special function groups */
            switch (op->groupId) {
            case SPECIAL_FUNCTION_SECURITY_GROUP: {
                me->security.localZoneId = op->localZoneId;
                me->security.groupId = op->groupId;
                me->security.key = op->key;
                me->security.fid = op->fid;
                me->security.opTimer1 = SelectA_GetSharedOpTimeEvt1(me, op->localZoneId, op->groupId, op->buttonId, 0);
                me->security.opTimer2 = SelectA_GetSharedOpTimeEvt2(me, op->localZoneId, op->groupId, op->buttonId, 0);
                me->security.deviceId = op->deviceId;
                Security_dispatch(&(me->security), e);
                break;
            }
            case SPECIAL_FUNCTION_ACCESS_GROUP: {
                me->access.localZoneId = op->localZoneId;
                me->access.groupId = op->groupId;
                me->access.key = op->key;
                me->access.fid = op->fid;
                me->access.absent = (me->apartmentState != APARTMENT_STATE_PRESENT) ? TRUE : FALSE;
                me->access.opTimer1 = SelectA_GetSharedOpTimeEvt1(me, op->localZoneId, op->groupId, op->buttonId, 0);
                me->access.opTimer2 = SelectA_GetSharedOpTimeEvt2(me, op->localZoneId, op->groupId, op->buttonId, 0);
                me->access.deviceId = op->deviceId;
                Access_dispatch(&(me->access), e);
                break;
            }
            }
            return;
        }
        if ((op->groupId > MAX_USER_GROUP_WITH_HSM) || (op->groupId < USER_GROUP)) {
            SelectA_UnhandledZoneGroupOperation(e);
            return;
        }
        groupIdx = op->groupId - USER_GROUP;
        state = &(me->clusterHsmStateOffsets[groupIdx]);
        if (*state == 0) {
            SelectA_UnhandledZoneGroupOperation(e);
            return;
        }
        stateMachineGroupId = 1;
    }
    /* specific group */
    switch (stateMachineGroupId) {
    /* light */
    case LIGHT_GROUP:
    case SHADE_GROUP:
    case HEATING_GROUP:
    case AUDIO_GROUP:
    case VIDEO_GROUP:
    case SECURITY_GROUP:
    case ACCESS_GROUP:
    case COOLING_GROUP:
    case VENTILATION_GROUP:
    case WINDOW_GROUP: {
        me->light.deviceId = op->deviceId;
        me->light.localZoneId = op->localZoneId;
        me->light.groupId = op->groupId;
        me->light.stateMachineGroupId = (op->groupId >= USER_GROUP? me->stateMachineGroupId[op->groupId-USER_GROUP] : op->groupId);
        me->light.buttonId = op->buttonId;
        me->light.apartmentWide = op->apartmentWide;
        me->light.outputChannel = op->outputChannel;
        me->light.opTimer1 = SelectA_GetSharedOpTimeEvt1(me, op->localZoneId, op->groupId, (me->progMode.localZoneId == 0? op->buttonId : 0), op->outputChannel);
        me->light.opTimer2 = SelectA_GetSharedOpTimeEvt2(me, op->localZoneId, op->groupId, (me->progMode.localZoneId == 0? op->buttonId : 0), op->outputChannel);
        me->light.stopTimer = SelectA_GetStopTimer(op->localZoneId, op->groupId, 0);
        me->light.super.state = GetStateFromStateHandlerOffset(*state);
        Light_dispatch(&(me->light), e);
        *state = GetStateHandlerOffsetFromState(me->light.super.state);
        break;
    }
    }
}

/*..........................................................................*/
static void SelectA_HandleShortUpstreamOperation(SelectA *me)
{

    if (me->key == KEY_DIRECT_SC) {
        if ((Datamodel_DeviceButtonId(&me->device) == BUTTON_SERVER)) {
            DsmApi_SendDeviceDirectActionEvent(me->device.deviceId, me->device.localZoneId, me->groupId, me->actionId);
            SelectA_SendGeneralAck();
            return;
        } else {
            SelectA_DirectSceneCall(me);
            SelectA_SendGeneralAck();
            return;
        }
    }

    QEvent e;
    Operation op;
    uint8_t buttonId = ((me->groupId <= MAX_GROUP) ? SelectA_GetButtonId(&me->device, me->key) : BUTTON_ZONE_0);

    if ((buttonId == BUTTON_SERVER) || ((me->groupId >= USER_GROUP) && (me->groupId <= MAX_GROUP))) {
        /* handle special function group IDs */
        if ((buttonId == BUTTON_SERVER) ||
            (me->groupId >= MIN_USER_GROUP_WITHOUT_HSM) ||
            (me->clusterHsmStateOffsets[me->groupId - USER_GROUP] == 0)) {
            /* no state machine -> send device action event, generalAck and exit */
            DsmApi_SendDeviceActionEvent(me->device.deviceId, me->device.localZoneId, me->groupId, me->key, me->click);
            SelectA_SendGeneralAck();
            return;
        }
    }
    /* get button ID of button/key */
    op.buttonId = buttonId;
    op.groupId = me->groupId;
    op.key = me->key;
    op.localZoneId = ((me->groupId <= MAX_GROUP) ? Datamodel_DeviceButtonZone(&me->device) : 0);
    op.fid = SelectA_GetFunctionId(&me->device);
    op.setLocalPriority = SelectA_GetLocalPriority(&me->device, me->key);
    e.dynamic_ = 0;
    op.outputChannel = me->device.buttonOutputChannel;
    e.sig = SelectA_GetSignal(me->click, &op);
    op.deviceId = me->device.deviceId;
    op.apartmentWide = me->device.buttonApartment;
    if (e.sig == CT_LOCAL_STOP_SIG) {
        /* send event */
        DsmApi_SendDeviceLocalActionEvent(me->device.deviceId, me->device.localZoneId, 2);
        /* send general ack */
        SelectA_SendGeneralAck();
    } else if (me->device.buttonNoComingHomeCall == 0) {
        if (op.groupId != SPECIAL_FUNCTION_ACCESS_GROUP) {
            if (me->apartmentState == APARTMENT_STATE_ABSENT) {
                /* send PRESENT scene to all devices on dSM API */
                DsmApi_SendZoneGroupActionCallSceneRequest(0, 0, PRESENT, me->device.deviceId, FALSE);
                /* send PRESENT scene to all devices on line */
                SelectA_SendSceneZoneGroup(0, 0, PRESENT, FALSE);
            }
            me->apartmentState = APARTMENT_STATE_PRESENT;
        }
    }
    if (e.sig == CT_LOCAL_ON_SIG || e.sig == CT_LOCAL_OFF_SIG) {
        /* send event */
        DsmApi_SendDeviceLocalActionEvent(me->device.deviceId, me->device.localZoneId, (e.sig == CT_LOCAL_ON_SIG) ? 1 : 0);
        /* no further action required */
        return;
    }
    if (e.sig == CT_SHORT_LONG_SIG && (Datamodel_DsmPropertiesProgrammingModeDisabled() || (op.groupId >= USER_GROUP))) {
        /* send general ack */
        SelectA_SendGeneralAck();
        /* no further action required */
        return;
    }
    SelectA_HandleOperation(me, &op, &e);
}

/*..........................................................................*/
static void SelectA_HandleSceneAllOperation(SelectA *me, SceneZoneGroupEvt const * e)
{
    Operation op;
    uint8_t sceneId = e->basic.sceneId;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = e->basic.outputChannel;
    switch (e->basic.super.sig) {
    case CALL_SCENE_BROADCAST_SIG: {
        /* update absent/present */
        switch (sceneId) {
        case PRESENT: {
            me->apartmentState = APARTMENT_STATE_PRESENT;
            break;
        }
        case ABSENT: {
            me->apartmentState = APARTMENT_STATE_ABSENT;
            break;
        }
        case PANIC:
        case ALARM1:
        case ALARM2:
        case ALARM3:
        case ALARM4:
        case FIRE: {
            me->alarmScene = sceneId;
            break;
        }
        }
        /* send call scene to line */
        SelectA_SendSceneZoneGroup(0, 0, (Scene)sceneId, e->basic.force);
        op.buttonId = Scene_GetButtonForScene(sceneId);
        break;
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        /* send undo scene number to line */
        SelectA_SendUndoSceneNumberZoneGroup(0, 0, (Scene)sceneId);
        op.buttonId = BUTTON_ZONE_0;
        break;
    }
    }
    for (op.localZoneId = MIN_ZONE; op.localZoneId <= MAX_ZONE; ++op.localZoneId) {
        for (op.groupId = MIN_GROUP; op.groupId <= MAX_USER_GROUP_WITH_HSM; ++op.groupId) {
            SelectA_HandleOperation(me, &op, (QEvent const *)e);
        }
    }
    /* handle special function groups */
    op.localZoneId = 0;
    op.buttonId = BUTTON_ZONE_0;
    for (op.groupId = MIN_SPECIAL_FUNCTION_GROUP; op.groupId <= MAX_SPECIAL_FUNCTION_GROUP; ++op.groupId) {
        SelectA_HandleOperation(me, &op, (QEvent const *)e);
    }
}

/*..........................................................................*/
static void SelectA_HandleSceneGroupOperation(SelectA *me, SceneZoneGroupEvt const * e)
{
    Operation op;
    uint8_t sceneId = e->basic.sceneId;
    op.groupId = e->groupId;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = e->basic.outputChannel;
    switch (e->basic.super.sig) {
    case CALL_SCENE_BROADCAST_SIG: {
        /* send call scene locally */
        SelectA_SendSceneZoneGroup(0, op.groupId, (Scene)sceneId, e->basic.force);
        op.buttonId = Scene_GetButtonForScene(sceneId);
        break;
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        /* send undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(0, op.groupId, (Scene)sceneId);
        op.buttonId = BUTTON_ZONE_0;
        break;
    }
    }
    /* handle main groups */
    for (op.localZoneId = MIN_ZONE; op.localZoneId <= MAX_ZONE; ++op.localZoneId) {
        SelectA_HandleOperation(me, &op, (QEvent const *)e);
    }
}

/*..........................................................................*/
static void SelectA_HandleSceneZoneOperation(SelectA *me, SceneZoneGroupEvt const * e)
{
    Operation op;
    uint8_t sceneId = e->basic.sceneId;
    op.localZoneId = e->localZoneId;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = e->basic.outputChannel;
    switch (e->basic.super.sig) {
    case CALL_SCENE_BROADCAST_SIG: {
        /* send call scene locally */
        SelectA_SendSceneZoneGroup(op.localZoneId, 0, (Scene)sceneId, e->basic.force);
        op.buttonId = Scene_GetButtonForScene(sceneId);
        break;
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        /* send undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(op.localZoneId, 0, (Scene)sceneId);
        op.buttonId = BUTTON_ZONE_0;
        break;
    }
    }
    /* handle main groups */
    for (op.groupId = MIN_GROUP; op.groupId <= MAX_USER_GROUP_WITH_HSM; ++op.groupId) {
        SelectA_HandleOperation(me, &op, (QEvent const *)e);
    }
}

/*..........................................................................*/
static void SelectA_HandleSceneZoneGroupOperation(SelectA *me, SceneZoneGroupEvt const * e)
{
    Operation op;
    op.localZoneId = e->localZoneId;
    /* handle main group */
    op.groupId = e->groupId;
    op.fid = 0;
    op.key = 0xff;
    op.buttonId = (e->basic.force || e->basic.super.sig == UNDO_SCENE_NUMBER_ZONE_GROUP_SIG) ? BUTTON_ZONE_0 : Scene_GetButtonForScene(e->basic.sceneId);
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = e->basic.outputChannel;
    SelectA_HandleOperation(me, &op, (QEvent const *)e);
}

/*..........................................................................*/
static void SelectA_HandleSceneDeviceOperation(SelectA *me, SceneDeviceEvt const * e)
{
    switch (e->super.sig) {
    case CALL_SCENE_DEVICE_SIG: {
        SelectASocket_SendSceneDevice(e->deviceId, (Scene)e->sceneId, e->force);
        break;
    }
    case UNDO_SCENE_NUMBER_DEVICE_SIG: {
        SelectASocket_SendUndoSceneNumberDevice(e->deviceId, (Scene)e->sceneId);
        break;
    }
    default: {
        break;
    }
    }
}

/*..........................................................................*/
static void SelectA_HandleSetLocalPrioAllOperation(SelectA *me, ActionZoneGroupSceneEvt const * e)
{
    Operation op;
    op.buttonId = Scene_GetButtonForScene(e->sceneId);
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = 0;

    /* set local priority on scene locally */
    SelectASocket_SendSetSceneLocalPrioZoneGroup(0, 0, (Scene)e->sceneId);

    if (Scene_IsOnScene(e->sceneId)) {
        for (op.localZoneId = MIN_ZONE; op.localZoneId <= MAX_ZONE; ++op.localZoneId) {
            for (op.groupId = MIN_GROUP; op.groupId <= MAX_USER_GROUP_WITH_HSM; ++op.groupId) {
                SelectA_HandleOperation(me, &op, (QEvent const *)e);
            }
        }
    }
}

/*..........................................................................*/
static void SelectA_HandleSetLocalPrioGroupOperation(SelectA *me, ActionZoneGroupSceneEvt const * e)
{
    Operation op;
    op.buttonId = Scene_GetButtonForScene(e->sceneId);
    op.groupId = e->groupId;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = 0;

    /* set local priority on scene locally */
    SelectASocket_SendSetSceneLocalPrioZoneGroup(0, e->groupId, (Scene)e->sceneId);

    if (Scene_IsOnScene(e->sceneId)) {
        for (op.localZoneId = MIN_ZONE; op.localZoneId <= MAX_ZONE; ++op.localZoneId) {
            SelectA_HandleOperation(me, &op, (QEvent const *)e);
        }
    }
}

/*..........................................................................*/
static void SelectA_HandleSetLocalPrioZoneOperation(SelectA *me, ActionZoneGroupSceneEvt const * e)
{
    Operation op;
    op.buttonId = Scene_GetButtonForScene(e->sceneId);
    op.localZoneId = e->localZoneId;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = 0;

    /* set local priority on scene locally */
    SelectASocket_SendSetSceneLocalPrioZoneGroup(e->localZoneId, 0, (Scene)e->sceneId);

    if (Scene_IsOnScene(e->sceneId)) {
        for (op.groupId = MIN_GROUP; op.groupId <= MAX_USER_GROUP_WITH_HSM; ++op.groupId) {
            SelectA_HandleOperation(me, &op, (QEvent const *)e);
        }
    }
}

/*..........................................................................*/
static void SelectA_HandleSetLocalPrioZoneGroupOperation(SelectA *me, ActionZoneGroupSceneEvt const * e)
{
    Operation op;
    op.buttonId = Scene_GetButtonForScene(e->sceneId);
    op.localZoneId = e->localZoneId;
    op.groupId = e->groupId;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = 0;

    /* set local priority on scene locally */
    SelectASocket_SendSetSceneLocalPrioZoneGroup(e->localZoneId, e->groupId, (Scene)e->sceneId);

    if (Scene_IsOnScene(e->sceneId)) {
        SelectA_HandleOperation(me, &op, (QEvent const *)e);
    }

}

/*..........................................................................*/
static void SelectA_HandleCallSceneMinAllOperation(SelectA *me, ActionZoneGroupSceneEvt const * e)
{
    /* call sceneMin */
    SelectASocket_CallSceneMinZoneGroup(e->localZoneId, e->groupId, (Scene)e->sceneId);
}

/*..........................................................................*/
static void SelectA_HandleRegisterLocalPrioZoneOperation(SelectA *me, ActionZoneGroupSceneEvt const * e)
{
    Operation op;
    op.buttonId = Scene_GetButtonForScene(e->sceneId);
    op.localZoneId = e->localZoneId;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = 0;

    if (Scene_IsOnScene(e->sceneId)) {
        for (op.groupId = MIN_GROUP; op.groupId <= MAX_USER_GROUP_WITH_HSM; ++op.groupId) {
            SelectA_HandleOperation(me, &op, (QEvent const *)e);
        }
    }
}

/*..........................................................................*/
static void SelectA_HandleRegisterLocalPrioZoneGroupOperation(SelectA *me, ActionZoneGroupSceneEvt const * e)
{
    Operation op;
    op.buttonId = Scene_GetButtonForScene(e->sceneId);
    op.localZoneId = e->localZoneId;
    op.groupId = e->groupId;
    op.deviceId = 0;
    op.apartmentWide = FALSE;
    op.outputChannel = 0;

    if (Scene_IsOnScene(e->sceneId)) {
        SelectA_HandleOperation(me, &op, (QEvent const *)e);
    }

}

/*..........................................................................*/
static void SelectA_HandleSharedTimer1Operation(SelectA *me, SharedOpTimeEvt const * opTimeEvt)
{
    QEvent e;
    Operation op;
    op.localZoneId = opTimeEvt->localZoneId;
    op.groupId = opTimeEvt->groupId;
    op.buttonId = opTimeEvt->buttonId;
    op.deviceId = opTimeEvt->deviceId;
    op.apartmentWide = opTimeEvt->apartmentWide;
    op.outputChannel = opTimeEvt->outputChannel;
    e.dynamic_ = 0;
    e.sig = opTimeEvt->sig;
    /* free shared time event */
    ((SharedOpTimeEvt *)opTimeEvt)->sig = 0;
    SelectA_HandleOperation(me, &op, &e);
}

/*..........................................................................*/
static void SelectA_HandleSharedTimer2Operation(SelectA *me, SharedOpTimeEvt const * opTimeEvt)
{
    QEvent e;
    Operation op;
    op.localZoneId = opTimeEvt->localZoneId;
    op.groupId = opTimeEvt->groupId;
    op.buttonId = opTimeEvt->buttonId;
    op.deviceId = opTimeEvt->deviceId;
    op.apartmentWide = opTimeEvt->apartmentWide;
    op.outputChannel = opTimeEvt->outputChannel;
    e.dynamic_ = 0;
    e.sig = opTimeEvt->sig;
    /* free shared time event */
    ((SharedOpTimeEvt *)opTimeEvt)->sig = 0;
    SelectA_HandleOperation(me, &op, &e);
}

/*..........................................................................*/
static void SelectA_HandleProgrammingModeSignal(SelectA *me, QEvent const * e)
{
    if (me->progMode.localZoneId >= MIN_ZONE) {
        if (me->progMode.localZoneId <= MAX_ZONE) {
            Operation op;
            op.deviceId = 0;
            op.localZoneId = me->progMode.localZoneId;
            op.groupId = me->progMode.groupId;
            op.buttonId = me->progMode.buttonId;
            op.apartmentWide = FALSE;
            op.outputChannel = 0;
            SelectA_HandleOperation(me, &op, e);
        } else {
            /* handle remote programming mode signal without zone */
            if (e->sig == SELECT_A_EXIT_PROG_MODE_TIMEOUT_SIG) {
                /* terminate programming mode */
                if (me->progMode.localZoneId == (MAX_ZONE + 1)) {
                    /* exit programming mode */
                    me->zoneIds[MAX_ZONE + 1] = 0;
                    me->progMode.localZoneId = 0;
                }
            }
        }
    }
}

/*..........................................................................*/
static void SelectA_HandleDatamodelDeviceEntryResponse(SelectA *me, DatamodelResDeviceEntryEvt const *deviceEntryEvt)
{
    if (me->progMode.localZoneId != 0) {
        Operation op;
        op.localZoneId = me->progMode.localZoneId;
        if (op.localZoneId == deviceEntryEvt->device.localZoneId) {
            op.deviceId = 0;
            op.groupId = me->progMode.groupId;
            op.buttonId = me->progMode.buttonId;
            op.apartmentWide = FALSE;
            op.outputChannel = 0;
            SelectA_HandleOperation(me, &op, (QEvent const *)deviceEntryEvt);
        }
    }
}

/*..........................................................................*/
static uint8_t SelectA_GetTransactionId(SelectA *me)
{
    return me->transactionId;
}

/*..........................................................................*/
static void SelectA_InvalidateCaches(SelectA *me)
{
    /* clear cached device */
    memset(&me->device, 0, sizeof(me->device));
}

/*..........................................................................*/
static void SelectA_UpdateUserGroup(SelectA *me, uint8_t groupId, uint8_t stateMachineGroupId)
{
    uint8_t groupIdx;
    Q_ASSERT((groupId >= USER_GROUP) && (groupId <= MAX_USER_GROUP_WITH_HSM));
    groupIdx = groupId - USER_GROUP;
    /* save stateMachineGroupId for zone */
    me->stateMachineGroupId[groupIdx] = stateMachineGroupId;
    if (stateMachineGroupId) {
        /* initialize user group state machine (initial state) */
        me->clusterHsmStateOffsets[groupIdx] = me->lightGroupInitHsmStateOffset;
    } else {
        me->clusterHsmStateOffsets[groupIdx] = 0;
    }
}

/*..........................................................................*/
static bool SelectA_CheckStatelessScene(Scene scene)
{
    static const Scene statelessScenes[] = { DEC_S, INC_S, STOP_S,
                                             T1_DEC, T1_INC, T1_STOP_S, T1234_CONT,
                                             T2_DEC, T2_INC, T2_STOP_S,
                                             T3_DEC, T3_INC, T3_STOP_S,
                                             T4_DEC, T4_INC, T4_STOP_S, SIG_BELL, RAIN, NO_RAIN };
    uint8_t index;
    for (index = 0; index < Q_DIM(statelessScenes); ++index) {
        if (scene == statelessScenes[index]) {
            return FALSE;
        }
    }
    return TRUE;
}

/*..........................................................................*/
static void SelectA_SaveCallSceneZoneGroupHistory(SelectA * selectA, uint8_t localZoneId, uint8_t groupId, Scene scene, uint8_t flags)
{
    if (localZoneId == 0) {
        for (localZoneId = MIN_ZONE; localZoneId <= MAX_ZONE; ++localZoneId) {
            SelectA_SaveCallSceneZoneGroupHistory(selectA, localZoneId, groupId, scene, flags);
        }
    } else {
        ZoneSceneHistoryRingbuffer * rb = &(selectA->zoneSceneHistoryRingbuffer[localZoneId-1]);
        uint8_t index = rb->index;
        if ((rb->groupId[index] == (groupId | flags)) && (rb->scene[index] == scene)) {
            return;
        }
        /* increment index before saving scene */
        if (++index >= MAX_ZONE_SCENE_HISTORY) {
            index = 0;
        }
        rb->index = index;
        if (rb->count < MAX_ZONE_SCENE_HISTORY) {
            ++rb->count;
        }
        rb->groupId[index] = groupId | flags;
        rb->scene[index] = scene;
    }
}

/*..........................................................................*/
QSignal SelectA_GetSignal(uint8_t clickType, const Operation *op)
{
    if (op->key == KEY_2WAY_B || op->key == KEY_LOCAL_2WAY_B) {
        /* 2-way button up, local button has same behaviour */
        if (clickType < Q_DIM(l_2WayUpButtonKeyToSignal)) {
            return l_2WayUpButtonKeyToSignal[clickType];
        }
    } else if (op->key == KEY_2WAY_A) {
        /* 2-way button down */
        if (clickType < Q_DIM(l_2WayDownButtonKeyToSignal)) {
            return l_2WayDownButtonKeyToSignal[clickType];
        }
    } else if (op->key == KEY_LOCAL_2WAY_A) {
        /* local 2-way button down */
        if (clickType < Q_DIM(l_local2WayDownButtonKeyToSignal)) {
            return l_local2WayDownButtonKeyToSignal[clickType];
        }
    }
    if (op->buttonId != BUTTON_LOCAL) {
        /* any button except local button */
        if (clickType < Q_DIM(l_keyToSignal)) {
            return l_keyToSignal[clickType];
        }
    } else {
        /* local button */
        if (clickType < Q_DIM(l_localKeyToSignal)) {
            return l_localKeyToSignal[clickType];
        }
    }
    return CT_INVALID_SIG;
}

/*..........................................................................*/
static uint8_t SelectA_GetGroupId(DatamodelDevice const *device, uint8_t key)
{
    uint8_t groupId;
    switch (FidGetClass(device->fid)) {
    case FID_CLASS_SECURITY:
        switch (FidGetSubclass(device->fid)) {
        case FID_SUBCLASS_SECURITY_ALARM:
            return SPECIAL_FUNCTION_SECURITY_GROUP;
        }
        break;
    case FID_CLASS_ACCESS:
        switch (FidGetSubclass(device->fid)) {
        case FID_SUBCLASS_ACCESS_BELL:
        case FID_SUBCLASS_ACCESS_LEAVING:
        case FID_SUBCLASS_ACCESS_LOCKING:
            return SPECIAL_FUNCTION_ACCESS_GROUP;
        }
        break;
    case FID_CLASS_JOKER:
        if (Datamodel_DeviceButtonGroup(device) == JOKER_GROUP) {
            switch (Datamodel_DeviceButtonId(device)) {
            case JOKER_BUTTON_ALARM:
            case JOKER_BUTTON_PANIC:
                return SPECIAL_FUNCTION_SECURITY_GROUP;
            case JOKER_BUTTON_LEAVEHOME:
            case JOKER_BUTTON_DOORBELL:
                return SPECIAL_FUNCTION_ACCESS_GROUP;
            }
        }
        break;
    }
    groupId = Datamodel_DeviceButtonActiveGroup(device);

    if (key == 4) {
        /* local button */
        if (groupId >= USER_GROUP) {
            /* return user group */
        } else if (Fid_IsJocker(device->fid)) {
            /* initialize group ID with value saved in button ID and group configuration */
            groupId = Datamodel_DeviceButtonGroup(device);
        } else {
            groupId = FidGetClass(device->fid);
        }
    }
    return groupId;
}

/*..........................................................................*/
static uint8_t SelectA_GetButtonId(DatamodelDevice const *device, uint8_t key)
{
    if (key != 4 && key <= 15) {
        /* get button ID and group configuration */
        return Datamodel_DeviceButtonId(device);
    }
    Q_ASSERT(key == 4);
    /* local button ID */
    return BUTTON_LOCAL;
}

/*..........................................................................*/
static bool SelectA_GetLocalPriority(DatamodelDevice const *device, uint8_t key)
{
    if (key < 4) {
        /* get button zone */
        return device->buttonSetLocalPriority;
    }
    /* use device's zone as default */
    return FALSE;
}

/*..........................................................................*/
static uint16_t SelectA_GetFunctionId(DatamodelDevice const *device)
{
    switch (FidGetClass(device->fid)) {
    case FID_CLASS_JOKER:
        switch (Datamodel_DeviceButtonId(device)) {
        case JOKER_BUTTON_ALARM:
            return (FID_CLASS_SECURITY << 12) | (FID_SUBCLASS_SECURITY_ALARM << 6) | FM_ALARM_BUTTON_ALARM;
        case JOKER_BUTTON_PANIC:
            return (FID_CLASS_SECURITY << 12) | (FID_SUBCLASS_SECURITY_ALARM << 6) | FM_ALARM_BUTTON_PANIC;
        case JOKER_BUTTON_LEAVEHOME:
            return (FID_CLASS_ACCESS << 12) | (FID_SUBCLASS_ACCESS_LEAVING << 6);
        case JOKER_BUTTON_DOORBELL:
            return (FID_CLASS_ACCESS << 12) | (FID_SUBCLASS_ACCESS_BELL << 6);
        }
    }
    return device->fid;
}

/*..........................................................................*/
static uint8_t SelectA_GetLocalZoneId(SelectA *me, uint16_t zoneId)
{
    uint8_t localZoneId;
    for (localZoneId = 0; localZoneId <= MAX_ZONE; ++localZoneId) {
        if (me->zoneIds[localZoneId] == zoneId) {
            return localZoneId;
        }
    }
    return 0;
}
/*--------------------------------------------------------------------------*/
/* Sensors -----------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

typedef struct {
    uint8_t sensorType;
    uint8_t groupId;
} CachedSensor_t;

/* assignment of cached sensor to array indexes */
static const CachedSensor_t l_cachedSensorTypes[MAX_CACHED_SENSORS] = {
        {SENSOR_TYPE_ROOM_TEMPERATURE, BROADCAST_GROUP},
        {SENSOR_TYPE_ROOM_TEMPERATURE_REQUESTED, ROOM_TEMPERATURE_CONTROL_GROUP},
        {SENSOR_TYPE_ROOM_TEMPERATURE_CONTROL_VALUE, ROOM_TEMPERATURE_CONTROL_GROUP},
};

static void SelectA_SensorsInit(Sensors *sensors)
{
    /* initialize timestamps and values */
    uint8_t zoneIdx;
    uint8_t sensorIdx;
    for (zoneIdx = 0; zoneIdx < MAX_ZONES; ++zoneIdx) {
        for (sensorIdx = 0; sensorIdx < MAX_CACHED_SENSORS; ++sensorIdx) {
            sensors->timestamps[zoneIdx][sensorIdx] = 0;
            sensors->values[zoneIdx][sensorIdx] = SENSOR_INVALID_VALUE;
        }
    }
}

/*--------------------------------------------------------------------------*/
static void SelectA_SensorsInitZone(Sensors *sensors, uint8_t localZoneId)
{
    Q_ASSERT((localZoneId >= MIN_ZONE) && (localZoneId <= MAX_ZONE));
    uint8_t zoneIdx = localZoneId - MIN_ZONE;
    uint8_t sensorIdx;
    uint32_t time = SelectB_GetTimeSeconds();
    for (sensorIdx = 0; sensorIdx < MAX_CACHED_SENSORS; ++sensorIdx) {
        sensors->timestamps[zoneIdx][sensorIdx] = time;
        sensors->values[zoneIdx][sensorIdx] = SENSOR_INVALID_VALUE;
    }
}

/*--------------------------------------------------------------------------*/
static uint8_t SelectA_SensorsGetSensorIdx(uint8_t sensorType, uint8_t groupId)
{
    uint8_t sensorIdx;
    for (sensorIdx = 0; sensorIdx < MAX_CACHED_SENSORS; ++sensorIdx) {
        if ((l_cachedSensorTypes[sensorIdx].sensorType == sensorType) &&
           ((l_cachedSensorTypes[sensorIdx].groupId == groupId) || (groupId == DATAMODEL_GROUP_ID_NOT_APPLICABLE)))
        {
            return sensorIdx;
        }
    }
    return SENSOR_INVALID_INDEX;
}
/*--------------------------------------------------------------------------*/
static void SelectA_SensorsUpdateSensor(ZoneGroupSensorPushEvt const *reqEvt)
{
    if ((reqEvt->localZoneId >= MIN_ZONE) && (reqEvt->localZoneId <= MAX_ZONE)) {
        Sensors *sensors = &l_selectA.sensors;
        uint8_t zoneIdx = reqEvt->localZoneId - MIN_ZONE;
        uint8_t sensorIdx = SelectA_SensorsGetSensorIdx(reqEvt->sensorType, reqEvt->groupId);
        if (sensorIdx != SENSOR_INVALID_INDEX) {
            sensors->timestamps[zoneIdx][sensorIdx] = SelectB_GetTimeSeconds();
            sensors->values[zoneIdx][sensorIdx] = (reqEvt->sensorValue & 0xfff);
        }
    }
}

/*--------------------------------------------------------------------------*/
Sensor SelectA_SensorsGetSensor(uint8_t localZoneId, uint8_t sensorType)
{
    Q_ASSERT((localZoneId >= MIN_ZONE) && (localZoneId <= MAX_ZONE));
    Sensors *sensors = &l_selectA.sensors;
    Sensor sensor;
    uint8_t zoneIdx = localZoneId - MIN_ZONE;
    uint8_t sensorIdx = SelectA_SensorsGetSensorIdx(sensorType, DATAMODEL_GROUP_ID_NOT_APPLICABLE);
    if (sensorIdx != SENSOR_INVALID_INDEX) {
        uint32_t timestamp = sensors->timestamps[zoneIdx][sensorIdx];
        sensor.timestamp = timestamp;
        sensor.value = sensors->values[zoneIdx][sensorIdx];
    } else {
        sensor.timestamp = 0;
        sensor.value = SENSOR_INVALID_VALUE;
    }
    return sensor;
}
/*--------------------------------------------------------------------------*/
uint32_t SelectA_SensorsGetAge(uint32_t timestamp)
{
    uint32_t time = SelectB_GetTimeSeconds();
    return (timestamp < time) ? (time - timestamp) : 0;
}
