/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Enrique Robledo Arnuncio, for digitalSTROM AG <era@tegulan.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_AVAHI

#include <stdlib.h>
#include <time.h>
#include <utlist.h>

#include <avahi-common/domain.h>

#include <digitalSTROM/dsuid.h>

#include "coordination.h"
#include "role.h"
#include "util.h"

typedef struct announced_vdsm {
    struct announced_vdsm *next;
    char *avahi_service_name;
    bool resolved;
    bool collecting;
} announced_vdsm_t;

static bool g_initial_vdsm_browse_complete = false;
static bool g_initial_vdsm_resolver_complete = false;

/* List of known foreign vdSMs. Access to this list always happens in the avahi
 * service thread, so it does not require explicit mutual exclusion.
 */
static announced_vdsm_t* g_vdsm_list = NULL;
static int g_vdsm_conflicts_found = 0;
static bool g_collection_enabled = false;
static collection_initialized_callback_t g_collection_initialized_callback = NULL;

static announced_vdsm_t *announced_vdsm_new(const char* service_name)
{
    size_t name_mem_size = strlen(service_name) + 1;
    if (name_mem_size > AVAHI_LABEL_MAX) {
        vdsm_report(LOG_ERR, "Invalid service name. vdSM ignored: %s\n",
                    service_name);
        return NULL;
    }
    announced_vdsm_t * res = malloc(sizeof(announced_vdsm_t));
    if (res == NULL) {
        vdsm_report(LOG_ERR,
                    "Could not allocate memory for announced_vdsm: %s\n",
                    service_name);
        return NULL;
    }

    memset(res, 0, sizeof(announced_vdsm_t));
    res->avahi_service_name = malloc(name_mem_size);
    if (res->avahi_service_name == NULL) {
        free(res);
        vdsm_report(LOG_ERR, "Could not allocate memory for service name: %s\n",
                    service_name);
        return NULL;
    }
    strcpy(res->avahi_service_name, service_name);
    return res;
}

static void announced_vdsm_free(announced_vdsm_t *vdsm)
{
    free(vdsm->avahi_service_name);
    free(vdsm);
}

static void update_coordination_state(bool initial_resolver_completion)
{
    bool found_foreign_collecting_vdsm = false;
    announced_vdsm_t *foreign_vdsm = NULL;
    LL_FOREACH(g_vdsm_list, foreign_vdsm) {
        if (foreign_vdsm->collecting) {
            found_foreign_collecting_vdsm = true;
            break;
        }
    }
    if (initial_resolver_completion) {
        g_initial_vdsm_resolver_complete = true;
        vdsm_report(LOG_INFO,
                    "Collection initialization complete with %d vdSM conflict(s).\n",
                    g_vdsm_conflicts_found);
        if (g_collection_initialized_callback) {
            const bool collection_enabled = (g_vdsm_conflicts_found == 0);
            g_collection_initialized_callback(collection_enabled);
        }
    }

    const bool collection_previously_enabled = g_collection_enabled;
    if (!g_initial_vdsm_resolver_complete) {
        g_collection_enabled = false;
    } else {
        g_collection_enabled = !found_foreign_collecting_vdsm;
    }
    if (initial_resolver_completion ||
        g_collection_enabled != collection_previously_enabled) {
        const char* state = (g_collection_enabled) ? "ENABLED" : "RESTRICTED";
        vdsm_report(LOG_INFO, "vDC host collection: %s\n", state);
    }
}

void vdsm_coordination_set_collection_initialized_callback(
    collection_initialized_callback_t cb)
{
    g_collection_initialized_callback = cb;
}

bool vdsm_coordination_is_collection_enabled()
{
    return g_collection_enabled;
}

void vdsm_coordination_add_foreign_vdsm(const char* service_name)
{
    vdsm_coordination_set_foreign_vdsm_role(service_name, NULL);
}

static bool unresolved_foreign_vdsms()
{
    bool found_unresolved_foreign_vdsm = false;
    announced_vdsm_t *foreign_vdsm = NULL;
    LL_FOREACH(g_vdsm_list, foreign_vdsm) {
        if (!foreign_vdsm->resolved) {
            found_unresolved_foreign_vdsm = true;
            break;
        }
    }
    return found_unresolved_foreign_vdsm;
}

void vdsm_coordination_set_foreign_vdsm_role(
    const char* service_name,
    const VDSMRole* role)
{
    announced_vdsm_t *vdsm = NULL;
    LL_FOREACH(g_vdsm_list, vdsm) {
        if (strcmp(vdsm->avahi_service_name, service_name) == 0) {
            break;
        }
    }
    if (vdsm == NULL) {
        vdsm = announced_vdsm_new(service_name);
        if (vdsm == NULL) {
            return;
        }
        LL_PREPEND(g_vdsm_list, vdsm);
    }
    bool initial_resolver_completion = false;
    if (role == NULL) {
        vdsm->resolved = false;
    } else {
        vdsm->resolved = true;
        vdsm->collecting = role->collecting;
        if (vdsm->collecting) {
            ++g_vdsm_conflicts_found;
        }
        if (!g_initial_vdsm_resolver_complete &&
            g_initial_vdsm_browse_complete &&
            !unresolved_foreign_vdsms()) {
            initial_resolver_completion = true;
        }
    }
    update_coordination_state(initial_resolver_completion);
}

void vdsm_coordination_remove_foreign_vdsm(const char* service_name)
{
    announced_vdsm_t *vdsm = NULL;
    LL_FOREACH(g_vdsm_list, vdsm) {
        if (strcmp(vdsm->avahi_service_name, service_name) == 0) {
            break;
        }
    }
    if (vdsm) {
        LL_DELETE(g_vdsm_list, vdsm);
        announced_vdsm_free(vdsm);
    } else {
        vdsm_report(LOG_ERR, "remove_vdsm: unknown instance, %s\n",
                    service_name);
    }
    update_coordination_state(false);
}

void vdsm_coordination_initial_browse_complete()
{
    g_initial_vdsm_browse_complete = true;
    const bool initial_resolver_completion = !unresolved_foreign_vdsms();
    update_coordination_state(initial_resolver_completion);
}

#endif  //HAVE_AVAHI
