/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of vdSM, imported from libdSvDC.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_AVAHI

#include <stdlib.h>
#include <syslog.h>
#include <avahi-common/alternative.h>
#include <utlist.h>

#include "discovery.h"

#include "coordination.h"
#include "role.h"
#include "util.h"
#include "vdc-api-proto.h"
#include "whitelist.h"

typedef struct discovered_vdc_hosts_list_element {
    struct discovered_vdc_host *next;
    const char* host;
    const char *port;
    dsuid_t dsuid;
    const char *service_name;
    bool noauto;
} discovered_vdc_hosts_list_element_t;

/* List of vdc hosts whose processing has been delayed because they have been
 * discovered during initialization. Access to this list always happens in the
 * avahi service thread, so it does not require explicit mutual exclusion.
 */
static discovered_vdc_hosts_list_element_t *g_discovered_vdc_hosts_list = NULL;

static bool g_collection_initialized = false;

static void vdsm_avahi_create_services(vdsm_avahi_handle_t *handle,
                                       AvahiClient *client);

static void vdsm_avahi_entry_group_callback(AvahiEntryGroup *group,
                                            AvahiEntryGroupState state,
                                            AVAHI_GCC_UNUSED void *userdata)
{
    if (userdata == NULL)
    {
        vdsm_report(LOG_ERR, "invalid userdata / missing handle\n");
        return;
    }

    vdsm_avahi_handle_t *handle = (vdsm_avahi_handle_t *)userdata;

    if (!(handle->group == group || group == NULL))
    {
        return; /* this was an assert in the avahi example */
    }

    switch (state)
    {
        case AVAHI_ENTRY_GROUP_ESTABLISHED:
            vdsm_report(LOG_DEBUG, "service \"%s\" successfully established.\n",
                   handle->name);
            break;

        case AVAHI_ENTRY_GROUP_COLLISION:
        {
            char *new_name;

            new_name = avahi_alternative_service_name(handle->name);
            if (handle->name)
            {
                avahi_free(handle->name);
            }
            handle->name = new_name;

            vdsm_report(LOG_WARNING,
                        "service name collision, renaming service to '%s'\n",
                        handle->name);
            /* recreate the service */
            vdsm_avahi_create_services(handle,
                                       avahi_entry_group_get_client(group));
            break;
        }
        case AVAHI_ENTRY_GROUP_FAILURE :
            vdsm_report(LOG_ERR, "entry group failure: %s\n",
                avahi_strerror(
                    avahi_client_errno(avahi_entry_group_get_client(group))));
            /* failure while registering service */
            avahi_simple_poll_quit(handle->simple_poll);
            break;

        case AVAHI_ENTRY_GROUP_UNCOMMITED:
        case AVAHI_ENTRY_GROUP_REGISTERING:
        default:
            break;
    }
}

static AvahiStringList* vdsm_avahi_create_txt_records(vdsm_avahi_handle_t *handle)
{
    const char *key = "dSUID=";
    size_t len = strlen(key) + strlen(handle->vdsm_dsuid) + 1;
    char *dsuid = malloc(len);
    if (!dsuid)
    {
        return NULL;
    }
    snprintf(dsuid, len, "%s%s", key, handle->vdsm_dsuid);
    AvahiStringList* list = avahi_string_list_new(dsuid, NULL);
    if (!list)
    {
        return NULL;
    }
    if (handle->vdsm_role.collectable)
    {
        list = avahi_string_list_add(list, "collectable");
        if (!list)
        {
            return NULL;
        }
    }

    if (handle->vdsm_role.collecting)
    {
        list = avahi_string_list_add(list, "collecting");
        if (!list)
        {
            return NULL;
        }
    }
    if (handle->vdsm_role.master)
    {
        list = avahi_string_list_add(list, "master");
        if (!list)
        {
            return NULL;
        }
    }
    if (handle->vdsm_role.noauto)
    {
        list = avahi_string_list_add(list, "noauto");
        if (!list)
        {
            return NULL;
        }
    }
    return list;
}

void vdsm_avahi_parse_role_txt_records(AvahiStringList *txt, VDSMRole* role)
{
    role->collectable = (avahi_string_list_find(txt, "collectable") != NULL);
    role->collecting = (avahi_string_list_find(txt, "collecting") != NULL);
    role->master = (avahi_string_list_find(txt, "master") != NULL);
    role->auxiliary = (avahi_string_list_find(txt, "auxiliary") != NULL);
    role->noauto = (avahi_string_list_find(txt, "noauto") != NULL);
}

static void vdsm_avahi_create_services(vdsm_avahi_handle_t *handle,
                                       AvahiClient *client)
{
    int ret = 0;
    char *new_name = NULL;

    if (!client)
    {
        return;
    }

    if (!handle)
    {
        return;
    }

    /* If this is the first time we're called, let's create a new
     * entry group if necessary */

    if (!handle->group)
    {
        handle->group = avahi_entry_group_new(client,
                                    vdsm_avahi_entry_group_callback, handle);
        if (!handle->group)
        {
            vdsm_report(LOG_ERR, "avahi_entry_group_new() failed: %s\n",
                                    avahi_strerror(avahi_client_errno(client)));
            goto fail;
        }
    }
    /* add our entries if the group is empty */
    if (avahi_entry_group_is_empty(handle->group))
    {
        vdsm_report(LOG_DEBUG, "Adding service '%s'\n", handle->name);

        AvahiStringList *txt = vdsm_avahi_create_txt_records(handle);
        if (!txt)
        {
            vdsm_report(LOG_ERR, "could not create _ds-vdsm._tcp txt records.\n");
            goto fail;
        }

        ret = avahi_entry_group_add_service_strlst(
            handle->group,
            AVAHI_IF_UNSPEC,
            AVAHI_PROTO_INET, 0,
            handle->name,
            "_ds-vdsm._tcp", NULL,
            NULL, handle->vdsm_port,
            txt);

        avahi_string_list_free(txt);

        if (ret < 0)
        {
            if (ret == AVAHI_ERR_COLLISION)
            {
                goto collision;
            }

            vdsm_report(LOG_ERR, "could not add _ds-vdsm._tcp service: %s\n",
                   avahi_strerror(ret));
            goto fail;
        }

        /* the service */
        ret = avahi_entry_group_commit(handle->group);
        if (ret < 0)
        {
            vdsm_report(LOG_ERR, "could not commit entry group: %s\n",
                   avahi_strerror(ret));
            goto fail;
        }
    }

    return;

collision:

    /* service name collision with a local service happened, pick a new name */
    new_name = avahi_alternative_service_name(handle->name);
    if (handle->name)
    {
        avahi_free(handle->name);
    }
    handle->name = new_name;

    vdsm_report(LOG_WARNING,
                "service name collision, renaming service to '%s'\n",
                handle->name);

    avahi_entry_group_reset(handle->group);

    vdsm_avahi_create_services(handle, client);
    return;

fail:
    avahi_simple_poll_quit(handle->simple_poll);
}

static void add_discovered_vdc_host(const char* host, const char *port,
                                    const dsuid_t* dsuid,
                                    const char *service_name, bool noauto);

static void remove_discovered_vdc_host(const char *service_name);

void vdsm_avahi_resolve_callback(
        AvahiServiceResolver *resolver,
        AVAHI_GCC_UNUSED AvahiIfIndex interface,
        AVAHI_GCC_UNUSED AvahiProtocol protocol,
        AvahiResolverEvent event,
        const char *name,
        const char *type,
        const char *domain,
        const char *host_name,
        const AvahiAddress *address,
        uint16_t port,
        AvahiStringList *txt,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        void* userdata)
{
    if (!resolver) {
        return;
    }

    /* callback is called whenever a service has been resolved successfully or
       timed out */

    switch (event)
    {
        case AVAHI_RESOLVER_FAILURE:
            vdsm_report(
                LOG_ERR,
                "failed to resolve service \"%s\" of type \"%s\" in domain \"%s\": %s\n",
                name, type, domain,
                avahi_strerror(
                    avahi_client_errno(
                        avahi_service_resolver_get_client(resolver))));
            break;
        case AVAHI_RESOLVER_FOUND:
        {
            char str_port[6]; /* "65535\0" */
            snprintf(str_port, 6, "%u", (unsigned)port);
            char ip[AVAHI_ADDRESS_STR_MAX];
            avahi_address_snprint(ip, sizeof(ip), address);
            char *dsuid_str = NULL;

            AvahiStringList *pair = avahi_string_list_find(txt, "dSUID");
            if (pair) {
                dsuid_str = strchr(avahi_string_list_get_text(pair), '=');
                if (dsuid_str) {
                    // skip '='
                    dsuid_str++;
                }
            }
            if (!dsuid_str) {
                vdsm_report(
                    LOG_ERR,
                    "ignoring %s %s at: %s:%s (%s) - missing dSUID in announcement\n",
                    type, name, host_name, str_port, ip);
                break;
            }
            dsuid_t dsuid;
            if (dsuid_from_string(dsuid_str, &dsuid) != DSUID_RC_OK) {
                vdsm_report(
                    LOG_ERR,
                    "ignoring %s %s at: %s:%s (%s) - invalid dSUID in announcement: %s\n",
                    type, name, host_name, str_port, ip, dsuid_str);
                break;
            }
            vdsm_report(LOG_DEBUG, "found %s at: %s:%s (%s)\n", name, host_name,
                        str_port, ip);
            if (strcmp(type, "_ds-vdc._tcp") == 0) {
                const AvahiStringList *noauto_slst =
                    avahi_string_list_find(txt, "noauto");
                const bool noauto = (noauto_slst != NULL);
                add_discovered_vdc_host(
                    host_name, str_port, &dsuid, name, noauto);
                if (g_collection_initialized) {
                    vdc_api_process_discovered_vdc_host(
                        host_name, str_port, &dsuid, name, noauto,
                        vdsm_coordination_is_collection_enabled());
                }
            } else if (strcmp(type, "_ds-vdsm._tcp") == 0) {
                if (!userdata) {
                    vdsm_report(LOG_ERR, "avahi: missing user data\n");
                } else {
                    vdsm_avahi_handle_t *avahi = userdata;
                    const bool is_self = (strcmp(avahi->name, name) == 0);
                    if (!is_self) {
                        VDSMRole role;
                        vdsm_avahi_parse_role_txt_records(txt, &role);
                        vdsm_coordination_set_foreign_vdsm_role(name, &role);
                    }
                }
            }
            break;
        }
        default:
            break;
    }

    avahi_service_resolver_free(resolver);
}

void vdsm_avahi_browse_callback(
        AvahiServiceBrowser *browser,
        AvahiIfIndex interface,
        AvahiProtocol protocol,
        AvahiBrowserEvent event,
        const char *name,
        const char *type,
        const char *domain,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        void* userdata)
{
    if (!browser) {
        vdsm_report(LOG_ERR, "avahi: invalid avahi service browser handle");
        return;
    }

    if (!userdata) {
        vdsm_report(LOG_ERR, "avahi: missing user data\n");
        return;
    }

    vdsm_avahi_handle_t *avahi = userdata;

    /* called when a new services becomes available or is removed from LAN */

    switch (event) {
        case AVAHI_BROWSER_FAILURE:

            vdsm_report(LOG_ERR, "avahi browse callback %s\n",
                   avahi_strerror(avahi_client_errno(
                           avahi_service_browser_get_client(browser))));
            avahi_simple_poll_quit(avahi->simple_poll);
            avahi->initialized = FALSE;
            return;

        case AVAHI_BROWSER_NEW:
            vdsm_report(
                LOG_DEBUG,
                "browser: new service \"%s\" of type \"%s\" in domain \"%s\"\n",
                name, type, domain);

            if (strcmp(type, "_ds-vdsm._tcp") == 0) {
                const bool is_self = (strcmp(avahi->name, name) == 0);
                if (!is_self) {
                    vdsm_report(LOG_INFO, "Foreign vdsm detected: %s\n", name);
                    vdsm_coordination_add_foreign_vdsm(name);
                }
            }
            /* We ignore the returned resolver object. In the callback
               function we free it. If the server is terminated before
               the callback function is called the server will free
               the resolver for us. */

            if (!(avahi_service_resolver_new(avahi->client,
                            interface, protocol, name, type, domain,
                            AVAHI_PROTO_INET, 0, vdsm_avahi_resolve_callback,
                            avahi))) {
                vdsm_report(LOG_ERR, "failed to resolve service \"%s\": %s\n",
                       name, avahi_strerror(avahi_client_errno(avahi->client)));
            }
            break;

        case AVAHI_BROWSER_REMOVE:
            vdsm_report(
                LOG_DEBUG,
                "browser: remove service \"%s\" of type \"%s\" in domain '%s'\n",
                name, type, domain);
            if (strcmp(type, "_ds-vdc._tcp") == 0) {
                vdc_api_remove_auto_connection(name);
                remove_discovered_vdc_host(name);
            } else if (strcmp(type, "_ds-vdsm._tcp") == 0) {
                vdsm_coordination_remove_foreign_vdsm(name);
            }
            break;

        case AVAHI_BROWSER_ALL_FOR_NOW:
            vdsm_report(LOG_DEBUG, "%s browser: ALL_FOR_NOW\n", type);
            if (strcmp(type, "_ds-vdsm._tcp") == 0) {
                vdsm_coordination_initial_browse_complete();
            }
            break;

        case AVAHI_BROWSER_CACHE_EXHAUSTED:
            vdsm_report(LOG_DEBUG, "%s browser: CACHE_EXHAUSTED\n", type);
        default:
            break;
    }
}

void vdsm_avahi_client_callback(
        AvahiClient *client,
        AvahiClientState state,
        AVAHI_GCC_UNUSED void * userdata)
{
    if (!client) {
        vdsm_report(LOG_ERR, "avahi: invalid client handle\n");
        return;
    }

    if (!userdata) {
        vdsm_report(LOG_ERR, "avahi: missing user data\n");
        return;
    }

    vdsm_avahi_handle_t *handle = userdata;

    switch (state)
    {
        case AVAHI_CLIENT_S_RUNNING:
            /* the server has startup successfully and registered its host
               name on the network, create our services */
            if (vdsm_role_needs_announcement(&handle->vdsm_role)) {
                vdsm_avahi_create_services(handle, client);
            }
            break;

        case AVAHI_CLIENT_S_COLLISION:
            /* Let's drop our registered services. When the server is back
             * in AVAHI_SERVER_RUNNING state we will register them
             * again with the new host name. */
        case AVAHI_CLIENT_S_REGISTERING:
            /* The server records are now being established. This
             * might be caused by a host name change. We need to wait
             * for our own records to register until the host name is
             * properly esatblished. */

            if (handle->group)
            {
                avahi_entry_group_reset(handle->group);
            }
            break;

        case AVAHI_CLIENT_FAILURE:
            vdsm_report(LOG_ERR, "avahi server conection failure: %s\n",
                   avahi_strerror(avahi_client_errno(client)));
            avahi_simple_poll_quit(handle->simple_poll);
            break;

        case AVAHI_CLIENT_CONNECTING:
        default:
            break;
    }
}

void free_discovered_vdc_host_list_element(
    discovered_vdc_hosts_list_element_t* e)
{
    free((void*)e->host);
    free((void*)e->port);
    free((void*)e->service_name);
    free(e);
}


discovered_vdc_hosts_list_element_t* new_discovered_vdc_hosts_list_element(
    const char *host,
    const char *port,
    const dsuid_t *dsuid,
    const char *service_name,
    bool noauto)
{
    discovered_vdc_hosts_list_element_t* e =
        malloc(sizeof(discovered_vdc_hosts_list_element_t));
    if (e == NULL) {
        vdsm_report(LOG_ERR, "could not allocate memory for vdc host list\n");
        return NULL;
    }
    memset(e, 0, sizeof(discovered_vdc_hosts_list_element_t));

    e->host = strdup(host);
    if (!e->host) {
        vdsm_report(LOG_ERR, "could not allocate memory for vdc host name\n");
        free_discovered_vdc_host_list_element(e);
        return NULL;
    }

    e->port = strdup(port);
    if (!e->port) {
        vdsm_report(LOG_ERR, "could not allocate memory for vdc host port\n");
        free_discovered_vdc_host_list_element(e);
        return NULL;
    }

    e->service_name = strdup(service_name);
    if (!e->service_name) {
        vdsm_report(LOG_ERR,
                    "could not allocate memory for vdc host service name\n");
        free_discovered_vdc_host_list_element(e);
        return NULL;
    }
    e->dsuid = *dsuid;
    e->noauto = noauto;
    return e;
}

void update_discovered_vdc_hosts_list_element(
    discovered_vdc_hosts_list_element_t* e,
    const char *host,
    const char *port,
    const char *service_name,
    bool noauto)
{
    if (strcmp(e->host, host) != 0) {
        const char* old_name = e->host;
        e->host = strdup(host);
        if (!e->host) {
            vdsm_report(LOG_ERR, "could not allocate memory for vdc host name\n");
            e->host = old_name;
        } else {
            free((void*)old_name);
        }
    }
    if (strcmp(e->port, port) != 0) {
        const char* old_port = e->port;
        e->port = strdup(port);
        if (!e->port) {
            vdsm_report(LOG_ERR, "could not allocate memory for vdc host port\n");
            e->port = old_port;
        } else {
            free((void*)old_port);
        }
    }
    if (strcmp(e->service_name, service_name) != 0) {
        const char* old_service_name = e->service_name;
        e->service_name = strdup(service_name);
        if (!e->service_name) {
            vdsm_report(LOG_ERR,
                        "could not allocate memory for vdc host service name\n");
            e->service_name = old_service_name;
        } else {
            free((void*)old_service_name);
        }
    }
    e->noauto = noauto;
}

static void add_discovered_vdc_host(
    const char* host,
    const char *port,
    const dsuid_t* dsuid,
    const char *service_name,
    bool noauto)
{
    /* The host may have been already added if the current discovery has been
     * triggered by a service name change, and the the corresponding service
     * removal has not yet arrived. In such case, we update any fields that
     * have changed.
     */
    discovered_vdc_hosts_list_element_t* vdc_host = NULL;
    LL_FOREACH(g_discovered_vdc_hosts_list, vdc_host) {
        if (dsuid_equal(&vdc_host->dsuid, dsuid)) {
            update_discovered_vdc_hosts_list_element(
                vdc_host, host, port, service_name, noauto);
            return;
        }
    }

    vdc_host = new_discovered_vdc_hosts_list_element(
        host, port, dsuid, service_name, noauto);
    if (vdc_host == NULL) {
        return;
    }
    LL_PREPEND(g_discovered_vdc_hosts_list, vdc_host);
}

static void remove_discovered_vdc_host(const char *service_name)
{
    discovered_vdc_hosts_list_element_t* vdc_host = NULL;
    discovered_vdc_hosts_list_element_t* tmp = NULL;
    LL_FOREACH_SAFE(g_discovered_vdc_hosts_list, vdc_host, tmp) {
        if (strcmp(vdc_host->service_name, service_name) == 0) {
            LL_DELETE(g_discovered_vdc_hosts_list, vdc_host);
            free_discovered_vdc_host_list_element(vdc_host);
        }
    }
}

void vdsm_collection_initialized_callback(bool collection_enabled)
{
    if (g_collection_initialized) {
        return;
    }
    discovered_vdc_hosts_list_element_t *vdc_host = NULL;
    LL_FOREACH(g_discovered_vdc_hosts_list, vdc_host) {
        vdsm_report(LOG_INFO, "Considering ignored client: %s\n",
                    vdc_host->service_name);
        vdc_api_process_discovered_vdc_host(
            vdc_host->host,
            vdc_host->port,
            &vdc_host->dsuid,
            vdc_host->service_name,
            vdc_host->noauto,
            collection_enabled);
    }
    g_collection_initialized = true;
}

void vdsm_whitelist_changed_callback()
{
    if (!g_collection_initialized) {
        return;
    }
    vdsm_report(LOG_DEBUG,
                "Whitelist changed. Reprocessing discovered vDC hosts.\n");
    char dsuid_str[DSUID_STR_LEN];
    discovered_vdc_hosts_list_element_t *vdc_host = NULL;
    LL_FOREACH(g_discovered_vdc_hosts_list, vdc_host) {
        if (vdc_api_in_client_list(&vdc_host->dsuid)) {
            // We are connected or attempting to connect to vdc_host. Its
            // autoconnect flag in the whitelist will be true unless the
            // whitelist has been changed externally, in which case we
            // need to disconnect.
            dsuid_to_string_upper(&vdc_host->dsuid, dsuid_str);
            const bool autoconnect = whitelist_get_autoconnect(dsuid_str);
            if (!autoconnect) {
                vdsm_report(
                    LOG_DEBUG,
                    "Autoconnect not set for %s in the whitelist.\n",
                    dsuid_str);
                // It would be somewhat more reliable to do this by dsuid
                // instead of by service_name
                vdc_api_remove_auto_connection(vdc_host->service_name);
            }
        } else {
            // We are not connected or attempting to connect to vdc_host.
            // Its autoconnect flag in the whitelist will be false unless the
            // whitelist has been changed externally.
            // vdc_api_process_discovered_vdc_host will check the autoconnect
            // flag, and try to establish a connection if it is set to true.
            vdc_api_process_discovered_vdc_host(
                vdc_host->host,
                vdc_host->port,
                &vdc_host->dsuid,
                vdc_host->service_name,
                vdc_host->noauto,
                vdsm_coordination_is_collection_enabled());
        }
    }
    vdsm_report(LOG_DEBUG, "Done reprocessing discovered vDC hosts.\n");
}

#endif//HAVE_AVAHI
