/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Enrique Robledo Arnuncio, for digitalSTROM AG <era@tegulan.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "whitelist.h"

#include <unistd.h>
#include <uthash.h>

#include <digitalSTROM/dsuid.h>

#include <algorithm>
#include <unordered_set>
#include <string>

using std::unordered_set;
using std::string;

string random_dsuid() {
    dsuid_t random_dsuid;
    char random_dsuid_str[DSUID_STR_LEN];
    dsuid_generate_v4_random(&random_dsuid);
    dsuid_to_string(&random_dsuid, random_dsuid_str);
    return string(random_dsuid_str);
}

string add_random_dsuid(unordered_set<string>* dsuids) {
    string dsuid;
    bool inserted = false;
    while (!inserted) {
        dsuid = random_dsuid();
        inserted = dsuids->insert(dsuid).second;
    }
    return dsuid;
}

unordered_set<string> random_dsuids(size_t count) {
    unordered_set<string> dsuids;
    dsuids.reserve(count);
    while (dsuids.size() < count)
        dsuids.insert(random_dsuid());
    return dsuids;
}

void create_old_whitelist(const unordered_set<string>& dsuids,
                          vdc_legacy_whitelist_hash_t **whitelist) {
    vdc_legacy_whitelist_hash_t* hash_entry = NULL;
    // for (const string& dsuid : dsuids) { // not supported in gcc 4.5
    for (auto itr = dsuids.begin(); itr != dsuids.end(); ++itr) {
        //printf("dsuid: %s\n", dsuid.c_str());
        hash_entry = (vdc_legacy_whitelist_hash_t*)
            malloc(sizeof(vdc_legacy_whitelist_hash_t));
        strncpy(hash_entry->dsuid, (*itr).data(), (*itr).size() + 1);
        HASH_ADD_STR(*whitelist, dsuid, hash_entry);
    }
}

static bool fail(const char* filename, unsigned lineno) {
    printf("Check failed on %s:%u\n", filename, lineno);
    exit(1);
    return false;
}

#define check(predicate) (predicate || fail(__FILE__, __LINE__))


int main() {
    char tmpdir_template[] = "/tmp/vdsm_whitelist_test_XXXXXX";
    const char* tmpdir = mkdtemp(tmpdir_template);
    if (tmpdir == NULL)
        return 1;
    const string whitelist_file = string(tmpdir) + "/test.db";
    check(whitelist_init(whitelist_file.c_str()));
    check(whitelist_exists());

    vdc_whitelist_fields_t fields;

    // Lookup on an empty whitelist
    check(!whitelist_get(random_dsuid().c_str(), &fields));

    vdc_legacy_whitelist_hash_t* whitelist = NULL;
    auto dsuids = random_dsuids(10);
    create_old_whitelist(dsuids, &whitelist);
    check(whitelist_import_legacy(&whitelist));

    // for (const string& dsuid : dsuids) { // not supported in gcc 4.5
    for (auto itr = dsuids.begin(); itr != dsuids.end(); ++itr) {
        check(whitelist_get((*itr).c_str(), &fields));
        // Imported devices are always registered with the following field values:
        check(fields.noauto == false);
        check(fields.autoconnect == true);
    }

    // Connection of devices in the whitelist without any changes.
    // for (const string& dsuid : dsuids) { // not supported in gcc 4.5
    for (auto itr = dsuids.begin(); itr != dsuids.end(); ++itr) {
        fields.noauto = false;       // Should be ignored by insert_and_get (unchanged)
        fields.autoconnect = false;  // Should be ignored by insert_and_get
        check(whitelist_insert_and_get_autoconnect((*itr).c_str(), &fields));
        check(whitelist_get((*itr).c_str(), &fields));
        check(fields.noauto == false);
        check(fields.autoconnect == true);
        check(fields.name == NULL);
    }

    // Connection of devices in the whitelist with noauto changed (to true)
    // for (const string& dsuid : dsuids) { // not supported in gcc 4.5
    for (auto itr = dsuids.begin(); itr != dsuids.end(); ++itr) {
        fields.noauto = true;        // Should be written by insert_and_get
        fields.autoconnect = false;  // Should be ignored by insert_and_get
        check(whitelist_insert_and_get_autoconnect((*itr).c_str(), &fields));
        check(whitelist_get((*itr).c_str(), &fields));
        check(fields.noauto == true);
        check(fields.autoconnect == true);
        check(fields.name == NULL);
        check(whitelist_get_autoconnect(itr->c_str()) == true);
    }

    // Connection of devices not in the whitelist.
    string new_dsuid = add_random_dsuid(&dsuids);
    check(!whitelist_get(new_dsuid.c_str(), &fields));  // Not there yet.
    fields.noauto = true;        // Should be written by insert_and_get
    fields.autoconnect = false;  // Should be written by insert_and_get
    check(!whitelist_insert_and_get_autoconnect(new_dsuid.c_str(), &fields));
    check(whitelist_get(new_dsuid.c_str(), &fields));
    check(fields.noauto == true);
    check(fields.autoconnect == false);
    check(fields.name == NULL);
    check(whitelist_get_autoconnect(new_dsuid.c_str()) == false);

    // Connection of devices in the whitelist with noauto changed (to false)
    fields.noauto = false;      // Should be written
    fields.autoconnect = true;  // Should be ignored
    check(!whitelist_insert_and_get_autoconnect(new_dsuid.c_str(), &fields));
    check(whitelist_get(new_dsuid.c_str(), &fields));
    check(fields.noauto == false);
    check(fields.autoconnect == false);
    check(fields.name == NULL);
    check(whitelist_get_autoconnect(new_dsuid.c_str()) == false);

    unlink(whitelist_file.c_str());
    rmdir(tmpdir);

    return 0;
}
